"""
Display 3D Greeks graphs

"""

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.axes3d import Axes3D # pylint: disable=unused-import
import numpy as np
import plotly.graph_objects as go
from plotly.offline import plot
# pylint: disable=invalid-name

class Greeks_3D():
    """
    Display 3D Greeks graphs

    """

    @staticmethod
    def graph_space_prep(params):
        """
        Prepare the axis ranges to be used in 3D graph.

        Parameters
        ----------
        greek : Str
            The sensitivity to be charted. Select from 'delta', 'gamma',
            'vega', 'theta', 'rho', 'vomma', 'vanna', 'zomma', 'speed',
            'color', 'ultima', 'vega_bleed', 'charm'. The default is
            'delta'
        S : Float
             Underlying Stock Price. The default is 100.
        axis : Str
            Whether the x-axis is 'price' or 'vol'. The default
            is 'price'.
        spacegrain : Int
            Number of points in each axis linspace argument for 3D
            graphs. The default is 100.

        Returns
        -------
        Various
            Updated parameters to be used in 3D graph.

        """

        graph_params = {}

        # Select the strike and Time ranges for each greek from the 3D
        # chart ranges dictionary
        SA_lower = params['3D_chart_ranges'][str(params['greek'])]['SA_lower']
        SA_upper = params['3D_chart_ranges'][str(params['greek'])]['SA_upper']
        TA_lower = params['3D_chart_ranges'][str(params['greek'])]['TA_lower']
        TA_upper = params['3D_chart_ranges'][str(params['greek'])]['TA_upper']

        # Set the volatility range from 5% to 50%
        sigmaA_lower = 0.05
        sigmaA_upper = 0.5

        # create arrays of 100 equally spaced points for the ranges of
        # strike prices, volatilities and maturities
        SA = np.linspace(SA_lower * params['S'],
                         SA_upper * params['S'],
                         int(params['spacegrain']))
        TA = np.linspace(TA_lower,
                         TA_upper,
                         int(params['spacegrain']))
        sigmaA = np.linspace(sigmaA_lower,
                             sigmaA_upper,
                             int(params['spacegrain']))

        # set y-min and y-max labels
        graph_params['ymin'] = TA_lower
        graph_params['ymax'] = TA_upper
        graph_params['axis_label2'] = 'Time to Expiration (Days)'

        # set x-min and x-max labels
        if params['axis'] == 'price':
            graph_params['x'], graph_params['y'] = np.meshgrid(SA, TA)
            graph_params['xmin'] = SA_lower
            graph_params['xmax'] = SA_upper
            graph_params['graph_scale'] = 1
            graph_params['axis_label1'] = 'Underlying Value'

        if params['axis'] == 'vol':
            graph_params['x'], graph_params['y'] = np.meshgrid(sigmaA, TA)
            graph_params['xmin'] = sigmaA_lower
            graph_params['xmax'] = sigmaA_upper
            graph_params['graph_scale'] = 100
            graph_params['axis_label1'] = 'Volatility %'

        return graph_params


    @classmethod
    def vis_greeks_3D(cls, graph_params, params):
        """
        Display 3D greeks graph.

        Returns
        -------
        If 'interactive' is False, a matplotlib static graph.
        If 'interactive' is True, a plotly graph that can be rotated
        and zoomed.

        """

        # Reverse the z-axis data if direction is 'short'
        if params['direction'] == 'short':
            graph_params['z'] = -graph_params['z']

        graph_params['titlename'] = cls._titlename(params=params)
        graph_params['axis_label3'] = str(params['greek'].title())

        if params['interactive']:
        # Create a plotly graph

            # Set the ranges for the contour values and reverse / rescale axes
            graph_params = cls._plotly_3D_ranges(graph_params=graph_params)

            return cls._plotly_3D(graph_params=graph_params, params=params)

        # Otherwise create a matplotlib graph
        graph_params = cls._mpl_axis_format(graph_params=graph_params)

        if params['gif']:
            fig, ax, titlename, title_font_scale = cls._mpl_3D(
                graph_params=graph_params, params=params)
            return fig, ax, titlename, title_font_scale

        return cls._mpl_3D(graph_params=graph_params, params=params)


    @staticmethod
    def _titlename(params):
        """
        Create graph title based on option type, direction and greek

        Returns
        -------
        Graph title.

        """

        titlename = str(str(params['direction'].title())
                        +' '
                        +params['option_title']
                        +' Option '
                        +str(params['greek'].title()))

        return titlename


    @staticmethod
    def _plotly_3D_ranges(graph_params):
        """
        Generate contour ranges and format axes for plotly 3D graph

        Returns
        -------
        axis ranges

        """
        # Set the ranges for the contour values and reverse / rescale axes
        graph_params['x'], graph_params['y'] = (
            graph_params['y'] * 365,
            graph_params['x'] * graph_params['graph_scale'])
        graph_params['x_start'] = graph_params['ymin']
        graph_params['x_stop'] = graph_params['ymax'] * 360
        graph_params['x_size'] = graph_params['x_stop'] / 18
        graph_params['y_start'] = graph_params['xmin']
        graph_params['y_stop'] = (
            graph_params['xmax'] * graph_params['graph_scale'])
        graph_params['y_size'] = (
            int((graph_params['xmax'] - graph_params['xmin']) / 20))
        graph_params['z_start'] = np.min(graph_params['z'])
        graph_params['z_stop'] = np.max(graph_params['z'])
        graph_params['z_size'] = (
            int((np.max(graph_params['z']) - np.min(graph_params['z'])) / 10))

        return graph_params


    @staticmethod
    def _plotly_3D(graph_params, params):
        """
        Display 3D greeks graph.

        Returns
        -------
        plotly 3D graph

        """
        # create plotly figure object
        fig = go.Figure(
            data=[go.Surface(x=graph_params['x'],
                             y=graph_params['y'],
                             z=graph_params['z'],

                             # set the colorscale to the chosen
                             # colorscheme
                             colorscale=params['colorscheme'],

                             # Define the contours
                             contours = {
                                 "x": {"show": True,
                                       "start": graph_params['x_start'],
                                       "end": graph_params['x_stop'],
                                       "size": graph_params['x_size'],
                                       "color":"white"},
                                 "y": {"show": True,
                                       "start": graph_params['y_start'],
                                       "end": graph_params['y_stop'],
                                       "size": graph_params['y_size'],
                                       "color":"white"},
                                 "z": {"show": True,
                                       "start": graph_params['z_start'],
                                       "end": graph_params['z_stop'],
                                       "size": graph_params['z_size']}
                                 },
                             )
                  ]
            )

        # Set initial view position
        camera = dict(
            eye=dict(x=2, y=1, z=1)
        )

        # Set x-axis to decrease from left to right
        fig.update_scenes(xaxis_autorange="reversed")

        # Set y-axis to increase from left to right
        fig.update_scenes(yaxis_autorange="reversed")
        fig.update_layout(scene = dict(
                            xaxis = dict(
                                 backgroundcolor="rgb(200, 200, 230)",
                                 gridcolor="white",
                                 showbackground=True,
                                 zerolinecolor="white",),
                            yaxis = dict(
                                backgroundcolor="rgb(230, 200,230)",
                                gridcolor="white",
                                showbackground=True,
                                zerolinecolor="white"),
                            zaxis = dict(
                                backgroundcolor="rgb(230, 230,200)",
                                gridcolor="white",
                                showbackground=True,
                                zerolinecolor="white",),
                            # Label axes
                            xaxis_title=graph_params['axis_label2'],
                            yaxis_title=graph_params['axis_label1'],
                            zaxis_title=graph_params['axis_label3'],),
                          title={'text':graph_params['titlename'],
                                 'y':0.9,
                                 'x':0.5,
                                 'xanchor':'center',
                                 'yanchor':'top',
                                 'font':dict(size=20,
                                             color="black")},
                          margin=dict(l=65, r=50, b=65, t=90),
                          scene_camera=camera)
        
        if params['web_graph'] is False:
            fig.update_layout(
                autosize=False,
                width=800,
                height=800
                )

        # If running in an iPython notebook the chart will display
        # in line
        if params['notebook']:
            # If output is sent to Dash
            if params['web_graph']:
                return fig

            fig.show()
            return

        # Otherwise create an HTML file that opens in a new window
        plot(fig, auto_open=True)
        return


    @staticmethod
    def _mpl_axis_format(graph_params):
        """
        Rescale Matplotlib axis values

        Returns
        -------
        x, y axis values

        """
        graph_params['x'] = graph_params['x'] * graph_params['graph_scale']
        graph_params['y'] = graph_params['y'] * 365

        return graph_params


    @staticmethod
    def _mpl_3D(params, graph_params):
        """
        Display 3D greeks graph.

        Returns
        -------
        Matplotlib static graph.

        """

        # Update chart parameters
        plt.style.use('seaborn-darkgrid')
        plt.rcParams.update(params['mpl_3d_params'])

        # create figure with specified size tuple
        fig = plt.figure(figsize=params['size3d'])
        ax = fig.add_subplot(111,
                             projection='3d',
                             azim=params['azim'],
                             elev=params['elev'])

        # Set background color to white
        ax.set_facecolor('w')

        # Create values that scale fonts with fig_size
        ax_font_scale = int(round(params['size3d'][0] * 1.1))
        title_font_scale = int(round(params['size3d'][0] * 1.8))

        # Tint the axis panes, RGB values from 0-1 and alpha denoting
        # color intensity
        ax.w_xaxis.set_pane_color((0.9, 0.8, 0.9, 0.8))
        ax.w_yaxis.set_pane_color((0.8, 0.8, 0.9, 0.8))
        ax.w_zaxis.set_pane_color((0.9, 0.9, 0.8, 0.8))

        # Set z-axis to left hand side
        ax.zaxis._axinfo['juggled'] = (1, 2, 0) # pylint: disable=protected-access

        # Set fontsize of axis ticks
        ax.tick_params(axis='both',
                       which='major',
                       labelsize=ax_font_scale,
                       pad=10)

        # Label axes
        ax.set_xlabel(graph_params['axis_label1'],
                      fontsize=ax_font_scale,
                      labelpad=ax_font_scale*1.5)
        ax.set_ylabel(graph_params['axis_label2'],
                      fontsize=ax_font_scale,
                      labelpad=ax_font_scale*1.5)
        ax.set_zlabel(graph_params['axis_label3'],
                      fontsize=ax_font_scale,
                      labelpad=ax_font_scale*1.5)

        # Auto scale the z-axis
        ax.set_zlim(auto=True)

        # Set x-axis to decrease from left to right
        ax.invert_xaxis()

        # apply graph_scale so that if volatility is the x-axis it
        # will be * 100
        ax.plot_surface(graph_params['x'],
                        graph_params['y'],
                        graph_params['z'],
                        rstride=2,
                        cstride=2,

                        # set the colormap to the chosen colorscheme
                        cmap=plt.get_cmap(params['colorscheme']),

                        # set the alpha value to the chosen
                        # colorintensity
                        alpha=params['colorintensity'],
                        linewidth=0.25)

        # Specify title
        st = fig.suptitle(graph_params['titlename'],
                          fontsize=title_font_scale,
                          fontweight=0,
                          color='black',
                          style='italic',
                          y=1.02)

        st.set_y(0.98)
        fig.subplots_adjust(top=1)

        if params['gif']:
            return fig, ax, graph_params['titlename'], title_font_scale

        # Display graph
        return plt.show()



"""
Utility functions for refreshing parameters

"""

import copy
import numpy as np
import scipy.stats as si
from optionvisualizer.visualizer_params import vis_params_dict
# pylint: disable=invalid-name

class Utils():
    """
    Utility functions for refreshing parameters

    """
    @staticmethod
    def _init_params(inputs):
        """
        Initialise parameter dictionary
        Parameters
        ----------
        inputs : Dict
            Dictionary of parameters supplied to the function.
        Returns
        -------
        params : Dict
            Dictionary of parameters.
        """
        # Copy the default parameters
        params = copy.deepcopy(vis_params_dict)

        # For all the supplied arguments
        for key, value in inputs.items():

            # Replace the default parameter with that provided
            params[key] = value

        return params


    @staticmethod
    def refresh_dist_params(opt_params, params):
        """
        Calculate various parameters and distributions

        Returns
        -------
        Various
            Assigns parameters to the object

        """

        # Cost of carry as risk free rate less dividend yield
        params['b'] = opt_params['r'] - opt_params['q']

        params['carry'] = np.exp(
            (params['b'] - opt_params['r']) * opt_params['T'])
        params['discount'] = np.exp(-opt_params['r'] * opt_params['T'])

        with np.errstate(divide='ignore'):
            params['d1'] = (
                (np.log(opt_params['S'] / opt_params['K'])
                 + (params['b'] + (0.5 * opt_params['sigma'] ** 2))
                 * opt_params['T'])
                / (opt_params['sigma'] * np.sqrt(opt_params['T'])))

            params['d2'] = (
                (np.log(opt_params['S'] / opt_params['K'])
                 + (params['b'] - (0.5 * opt_params['sigma'] ** 2))
                 * opt_params['T'])
                / (opt_params['sigma'] * np.sqrt(opt_params['T'])))

            # standardised normal density function
            params['nd1'] = (
                (1 / np.sqrt(2 * np.pi)) * (np.exp(-params['d1'] ** 2 * 0.5)))

            # Cumulative normal distribution function
            params['Nd1'] = si.norm.cdf(params['d1'], 0.0, 1.0)
            params['minusNd1'] = si.norm.cdf(-params['d1'], 0.0, 1.0)
            params['Nd2'] = si.norm.cdf(params['d2'], 0.0, 1.0)
            params['minusNd2'] = si.norm.cdf(-params['d2'], 0.0, 1.0)

        return params


    @staticmethod
    def refresh_combo_params(params, inputs):
        """
        Set parameters for use in various pricing functions

        Parameters
        ----------
        **kwargs : Various
                   Takes any of the arguments of the various methods
                   that use it to refresh data.

        Returns
        -------
        Various
            Runs methods to fix input parameters and reset defaults
            if no data provided

        """
        default_values = copy.deepcopy(vis_params_dict)

        # Certain combo payoffs (found in the mod_payoffs list) require
        # specific default parameters
        if params['combo_payoff'] in params['mod_payoffs']:

            # For each parameter in the combo parameters dictionary
            # corresponding to this combo payoff
            for parameter in params[
                    'combo_parameters'][params['combo_payoff']]:

                # If the parameter has been supplied as an input
                if parameter in inputs.keys():

                    # Set this value in the parameter dictionary
                    params[parameter] = inputs[parameter]

                # Otherwise if the parameter is in the mod_params list
                elif parameter in params['mod_params']:

                    # Try to extract this from the combo dict default
                    try:
                        params[parameter] = params['combo_dict'][str(
                            params['combo_payoff'])][str(parameter)]

                    # Otherwise
                    except KeyError:
                        # Set to the standard default value
                        params[parameter] = default_values[str(parameter)]
                # Otherwise
                else:
                    # Set to the standard default value
                    params[parameter] = default_values[str(parameter)]

        # For all the other combo_payoffs
        else:
            # For each parameter in the combo parameters dictionary
            # corresponding to this combo payoff
            for parameter in params[
                    'combo_parameters'][params['combo_payoff']]:

                # If the parameter has been supplied as an input
                if parameter in inputs.keys():

                    # Set this value in the parameter dictionary
                    params[parameter] = inputs[parameter]

                # Otherwise
                else:
                    # Set to the standard default value
                    params[parameter] = default_values[str(parameter)]

        return params


    @staticmethod
    def barrier_factors(params):
        """
        Calculate the barrier option specific parameters

        Returns
        -------
        Various
            Assigns parameters to the object

        """

        # Cost of carry as risk free rate less dividend yield
        params['b'] = params['r'] - params['q']

        barrier_factors = {}

        barrier_factors['mu'] = ((
            params['b'] - ((params['sigma'] ** 2) / 2))
            / (params['sigma'] ** 2))

        barrier_factors['lambda'] = (
            np.sqrt(barrier_factors['mu'] ** 2 + (
                (2 * params['r']) / params['sigma'] ** 2)))

        barrier_factors['z'] = (
            (np.log(params['H'] / params['S'])
             / (params['sigma'] * np.sqrt(params['T'])))
            + (barrier_factors['lambda']
               * params['sigma']
               * np.sqrt(params['T'])))

        barrier_factors['x1'] = (
            np.log(params['S'] / params['K'])
            / (params['sigma'] * np.sqrt(params['T']))
            + ((1 + barrier_factors['mu'])
               * params['sigma']
               * np.sqrt(params['T'])))

        barrier_factors['x2'] = (
            np.log(params['S'] / params['H'])
            / (params['sigma']
               * np.sqrt(params['T']))
            + ((1 + barrier_factors['mu'])
               * params['sigma']
               * np.sqrt(params['T'])))

        barrier_factors['y1'] = (
            np.log((params['H'] ** 2)
                   / (params['S']
                      * params['K']))
            / (params['sigma']
               * np.sqrt(params['T']))
            + ((1 + barrier_factors['mu'])
               * params['sigma']
               * np.sqrt(params['T'])))

        barrier_factors['y2'] = (
            np.log(params['H'] / params['S'])
            / (params['sigma']
               * np.sqrt(params['T']))
            + ((1 + barrier_factors['mu'])
               * params['sigma']
               * np.sqrt(params['T'])))

        params['carry'] = np.exp((params['b'] - params['r']) * params['T'])

        barrier_factors['A'] = (
            (params['phi']
             * params['S']
             * params['carry']
             * si.norm.cdf((params['phi']
                            * barrier_factors['x1']), 0.0, 1.0))
            - (params['phi']
               * params['K']
               * np.exp(-params['r']
                        * params['T'])
               * si.norm.cdf(
                   ((params['phi'] * barrier_factors['x1'])
                    - (params['phi'] * params['sigma']
                       * np.sqrt(params['T']))), 0.0, 1.0)))


        barrier_factors['B'] = (
            (params['phi']
             * params['S']
             * params['carry']
             * si.norm.cdf((params['phi']
                            * barrier_factors['x2']), 0.0, 1.0))
            - (params['phi']
               * params['K']
               * np.exp(-params['r']
                        * params['T'])
               * si.norm.cdf(
                   ((params['phi'] * barrier_factors['x2'])
                    - (params['phi'] * params['sigma']
                       * np.sqrt(params['T']))), 0.0, 1.0)))

        barrier_factors['C'] = (
            (params['phi']
             * params['S']
             * params['carry']
             * ((params['H'] / params['S'])
                ** (2 * (barrier_factors['mu'] + 1)))
             * si.norm.cdf((params['eta']
                            * barrier_factors['y1']), 0.0, 1.0))
            - (params['phi']
               * params['K']
               * np.exp(-params['r']
                        * params['T'])
               * ((params['H'] / params['S'])
                  ** (2 * barrier_factors['mu']))
               * si.norm.cdf(
                   ((params['eta'] * barrier_factors['y1'])
                    - (params['eta'] * params['sigma']
                       * np.sqrt(params['T']))), 0.0, 1.0)))

        barrier_factors['D'] = (
            (params['phi'] * params['S'] * params['carry']
             * ((params['H'] / params['S'])
                ** (2 * (barrier_factors['mu'] + 1)))
             * si.norm.cdf((params['eta']
                            * barrier_factors['y2']), 0.0, 1.0))
            - (params['phi']
               * params['K']
               * np.exp(-params['r']
                        * params['T'])
               * ((params['H'] / params['S'])
                  ** (2 * barrier_factors['mu']))
               * si.norm.cdf(
                   ((params['eta'] * barrier_factors['y2'])
                    - (params['eta'] * params['sigma']
                       * np.sqrt(params['T']))), 0.0, 1.0)))

        barrier_factors['E'] = (
            (params['R'] * np.exp(-params['r'] * params['T']))
            * (si.norm.cdf(
                ((params['eta'] * barrier_factors['x2'])
                 - (params['eta']
                    * params['sigma']
                    * np.sqrt(params['T']))), 0.0, 1.0)
                - (((params['H'] / params['S'])
                    ** (2 * barrier_factors['mu']))
                   * si.norm.cdf(
                       ((params['eta'] * barrier_factors['y2'])
                        - (params['eta'] * params['sigma']
                           * np.sqrt(params['T']))), 0.0, 1.0))))

        barrier_factors['F'] = (
            params['R'] * (((params['H'] / params['S'])
                            ** (barrier_factors['mu']
                                + barrier_factors['lambda']))
                      * (si.norm.cdf((params['eta']
                                      * barrier_factors['z']), 0.0, 1.0))
                      + (((params['H'] / params['S'])
                          ** (barrier_factors['mu']
                              - barrier_factors['lambda']))
                         * si.norm.cdf(
                             ((params['eta'] * barrier_factors['z'])
                              - (2 * params['eta']
                                 * barrier_factors['lambda']
                                 * params['sigma']
                                 * np.sqrt(params['T']))), 0.0, 1.0))))

        return barrier_factors, params



"""
Display 2D Greeks graphs

"""

from matplotlib import pylab
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.axes3d import Axes3D # pylint: disable=unused-import
import numpy as np
import plotly.graph_objects as go
from plotly.offline import plot
from optionvisualizer.sensitivities import Sens

# pylint: disable=invalid-name

class Greeks_2D():
    """
    Display 2D Greeks graphs

    """

    @classmethod
    def vis_greeks_2D(cls, params):
        """
        Creates data for 2D greeks graph.

        Returns
        -------
        Runs method to graph using Matplotlib.

        """

        # create arrays of 1000 equally spaced points for a range of
        # strike prices, volatilities and maturities
        params['SA'] = np.linspace(0.8 * params['S'], 1.2 * params['S'], 1000)
        params['sigmaA'] = np.linspace(0.05, 0.5, 1000)
        params['TA'] = np.linspace(0.01, 1, 1000)

        # y-axis parameters other than rho require 3 options to be
        # graphed
        if params['y_plot'] in params['y_name_dict'].keys():

            params = cls._non_rho_data(params)

        # rho requires 4 options to be graphed
        if params['y_plot'] == 'rho':

            params = cls._rho_data(params)

        # Convert the x-plot and y-plot values to axis labels
        xlabel = params['label_dict'][str(params['x_plot'])]
        ylabel = params['label_dict'][str(params['y_plot'])]

        # If the greek is rho or the same for a call or a put, set the
        # option name to 'Call / Put'
        if params['y_plot'] in [params['equal_greeks'], 'rho']:
            params['option'] = 'Call / Put'

        # Create chart title as direction plus option type plus y-plot
        # vs x-plot
        title = (str(params['direction'].title())
                 +' '
                 +str(params['option'].title())
                 +' '
                 +params['y_plot'].title()
                 +' vs '
                 +params['x_plot'].title())

        # Set the x-axis array as price, vol or time
        x_name = str(params['x_plot'])
        if x_name in params['x_name_dict'].keys():
            xarray = (params[str(params['x_name_dict'][x_name])] *
                      params['x_scale_dict'][x_name])

        vis_params = {
            'x_plot':params['x_plot'],
            'yarray1':params['C1'],
            'yarray2':params['C2'],
            'yarray3':params['C3'],
            'yarray4':params['C4'],
            'xarray':xarray,
            'label1':params['label1'],
            'label2':params['label2'],
            'label3':params['label3'],
            'label4':params['label4'],
            'xlabel':xlabel,
            'ylabel':ylabel,
            'title':title,
            'size2d':params['size2d'],
            'mpl_style':params['mpl_style'],
            'gif':params['gif']
            }

        # Plot 3 option charts
        if params['y_plot'] in params['y_name_dict'].keys():
            if params['interactive']:
                return cls._vis_greeks_plotly(
                    vis_params=vis_params, params=params)

            if params['gif']:
                fig, ax = cls._vis_greeks_mpl(
                    vis_params=vis_params, params=params)
                return fig, ax

            return cls._vis_greeks_mpl(
                vis_params=vis_params, params=params)

        # Plot Rho charts
        if params['y_plot'] == 'rho':
            if params['interactive']:
                return cls._vis_greeks_plotly(
                    vis_params=vis_params, params=params)

            vis_params.update({'gif':False})
            return cls._vis_greeks_mpl(
                vis_params=vis_params, params=params)

        return print("Please select a valid pair")


    @classmethod
    def _non_rho_data(cls, params):

        for opt in [1, 2, 3]:
            if params['x_plot'] == 'price':

                # For price we set S to the array SA
                params['C'+str(opt)] = Sens.sensitivities_static(
                    params=params, S=params['SA'],
                    K=params['G'+str(opt)], T=params['T'+str(opt)],
                    r=params['r'], q=params['q'], sigma=params['sigma'],
                    option=params['option'],
                    greek=params['y_name_dict'][params['y_plot']],
                    price_shift=0.25, vol_shift=0.001,
                    ttm_shift=(1 / 365), rate_shift=0.0001,
                    num_sens=params['num_sens'])

            if params['x_plot'] == 'vol':

                # For vol we set sigma to the array sigmaA
                params['C'+str(opt)] = Sens.sensitivities_static(
                    params=params, S=params['S'],
                    K=params['G'+str(opt)], T=params['T'+str(opt)],
                    r=params['r'], q=params['q'], sigma=params['sigmaA'],
                    option=params['option'],
                    greek=params['y_name_dict'][params['y_plot']],
                    price_shift=0.25, vol_shift=0.001,
                    ttm_shift=(1 / 365), rate_shift=0.0001)

            if params['x_plot'] == 'time':

                # For time we set T to the array TA
                params['C'+str(opt)] = Sens.sensitivities_static(
                    params=params, S=params['S'],
                    K=params['G'+str(opt)], T=params['TA'], r=params['r'],
                    q=params['q'], sigma=params['sigma'],
                    option=params['option'],
                    greek=params['y_name_dict'][params['y_plot']],
                    price_shift=0.25, vol_shift=0.001,
                    ttm_shift=(1 / 365), rate_shift=0.0001)

        # Reverse the option value if direction is 'short'
        if params['direction'] == 'short':
            for opt in [1, 2, 3]:
                params['C'+str(opt)] = -params['C'+str(opt)]

        # Call strike_tenor_label method to assign labels to chosen
        # strikes and tenors
        params = cls._strike_tenor_label(params)

        return params


    @staticmethod
    def _rho_data(params):

        # Set T1 and T2 to the specified time and shifted time
        params['T1'] = params['T']
        params['T2'] = params['T'] + params['time_shift']

        # 2 Tenors
        tenor_type = {1:1, 2:2, 3:1, 4:2}

        # And call and put for each tenor
        opt_type = {1:'call', 2:'call', 3:'put', 4:'put'}
        for opt in [1, 2, 3, 4]:
            if params['x_plot'] == 'price':

                # For price we set S to the array SA
                params['C'+str(opt)] = Sens.sensitivities_static(
                    params=params, S=params['SA'], K=params['G2'],
                    T=params['T'+str(tenor_type[opt])], r=params['r'],
                    q=params['q'], sigma=params['sigma'],
                    option=opt_type[opt], greek=params['y_plot'],
                    price_shift=0.25, vol_shift=0.001, ttm_shift=(1 / 365),
                    rate_shift=0.0001)

            if params['x_plot'] == 'strike':

                # For strike we set K to the array SA
                params['C'+str(opt)] = Sens.sensitivities_static(
                    params=params, S=params['S'], K=params['SA'],
                    T=params['T'+str(tenor_type[opt])], r=params['r'],
                    q=params['q'], sigma=params['sigma'],
                    option=opt_type[opt], greek=params['y_plot'],
                    price_shift=0.25, vol_shift=0.001, ttm_shift=(1 / 365),
                    rate_shift=0.0001)

            if params['x_plot'] == 'vol':

                # For vol we set sigma to the array sigmaA
                params['C'+str(opt)] = Sens.sensitivities_static(
                    params=params, S=params['S'], K=params['G2'],
                    T=params['T'+str(tenor_type[opt])], r=params['r'],
                    q=params['q'], sigma=params['sigmaA'],
                    option=opt_type[opt], greek=params['y_plot'],
                    price_shift=0.25, vol_shift=0.001, ttm_shift=(1 / 365),
                    rate_shift=0.0001)

        # Reverse the option value if direction is 'short'
        if params['direction'] == 'short':
            for opt in [1, 2, 3, 4]:
                params['C'+str(opt)] = -params['C'+str(opt)]

        # Assign the option labels
        params['label1'] = str(int(params['T1'] * 365))+' Day Call'
        params['label2'] = str(int(params['T2'] * 365))+' Day Call'
        params['label3'] = str(int(params['T1'] * 365))+' Day Put'
        params['label4'] = str(int(params['T2'] * 365))+' Day Put'

        return params


    @staticmethod
    def _strike_tenor_label(params):
        """
        Assign labels to chosen strikes and tenors in 2D greeks graph

        Returns
        -------
        Str
            Labels for each of the 3 options in 2D greeks graph.

        """
        strike_label = dict()
        for strike, strike_value in {'G1':'label1',
                                     'G2':'label2',
                                     'G3':'label3'}.items():

            # If the strike is 100% change name to 'ATM'
            if params[str(strike)] == params['S']:
                strike_label[strike_value] = 'ATM Strike'
            else:
                strike_label[strike_value] = (
                    str(int(params[strike]))
                    +' Strike')

        for tenor, tenor_value in {'T1':'label1',
                                   'T2':'label2',
                                   'T3':'label3'}.items():

            # Make each label value the number of days to maturity
            # plus the strike level
            params[tenor_value] = (
                str(int(params[str(tenor)]*365))
                +' Day '
                +strike_label[str(tenor_value)])

        return params


    @staticmethod
    def _vis_greeks_mpl(vis_params, params):
        """
        Display the 2D greeks chart using matplotlib

        Parameters
        ----------
        xarray : Array
            x-axis values.
        yarray1 : Array
            y-axis values for option 1.
        yarray2 : Array
            y-axis values for option 2.
        yarray3 : Array
            y-axis values for option 3.
        yarray4 : Array
            y-axis values for option 4.
        label1 : Str
            Option 1 label.
        label2 : Str
            Option 2 label.
        label3 : Str
            Option 3 label.
        label4 : Str
            Option 4 label.
        xlabel : Str
            x-axis label.
        ylabel : Str
            y-axis label.
        title : Str
            Chart title.

        Returns
        -------
        2D Greeks chart.

        """

        # Set style to chosen mpl_style (default is Seaborn Darkgrid)
        plt.style.use(vis_params['mpl_style'])

        # Update chart parameters
        pylab.rcParams.update(params['mpl_params'])

        # Create the figure and axes objects
        fig, ax = plt.subplots(figsize=vis_params['size2d'])

        # If plotting against time, show time to maturity reducing left
        # to right
        if vis_params['x_plot'] == 'time':
            ax.invert_xaxis()

        # Plot the 1st option
        ax.plot(vis_params['xarray'],
                vis_params['yarray1'],
                color='blue',
                label=vis_params['label1'])

        # Plot the 2nd option
        ax.plot(vis_params['xarray'],
                vis_params['yarray2'],
                color='red',
                label=vis_params['label2'])

        # Plot the 3rd option
        ax.plot(vis_params['xarray'],
                vis_params['yarray3'],
                color='green',
                label=vis_params['label3'])

        # 4th option only used in Rho graphs
        if vis_params['label4'] is not None:
            ax.plot(vis_params['xarray'],
                    vis_params['yarray4'],
                    color='orange',
                    label=vis_params['label4'])

        # Apply a grid
        plt.grid(True)

        # Apply a black border to the chart
        ax.patch.set_edgecolor('black')
        ax.patch.set_linewidth('1')

        # Set x and y axis labels and title
        ax.set(xlabel=vis_params['xlabel'],
               ylabel=vis_params['ylabel'],
               title=vis_params['title'])

        # Create a legend
        ax.legend(loc=0, fontsize=10)

        if vis_params['gif']:
            return fig, ax

            # Display the chart
        return plt.show()


    @classmethod
    def _vis_greeks_plotly(cls, vis_params, params):
        """
        Display the 2D greeks chart using plotly

        Parameters
        ----------
        xarray : Array
            x-axis values.
        yarray1 : Array
            y-axis values for option 1.
        yarray2 : Array
            y-axis values for option 2.
        yarray3 : Array
            y-axis values for option 3.
        yarray4 : Array
            y-axis values for option 4.
        label1 : Str
            Option 1 label.
        label2 : Str
            Option 2 label.
        label3 : Str
            Option 3 label.
        label4 : Str
            Option 4 label.
        xlabel : Str
            x-axis label.
        ylabel : Str
            y-axis label.
        title : Str
            Chart title.

        Returns
        -------
        2D Greeks chart.

        """

        # Create the figure
        fig = go.Figure()

        # If plotting against time, show time to maturity reducing left
        # to right
        if vis_params['x_plot'] == 'time':
            fig.update_xaxes(autorange="reversed")

        # Plot the 1st option
        fig.add_trace(go.Scatter(x=vis_params['xarray'],
                                 y=vis_params['yarray1'],
                                 line=dict(color='blue'),
                                 name=vis_params['label1']))

        # Plot the 2nd option
        fig.add_trace(go.Scatter(x=vis_params['xarray'],
                                 y=vis_params['yarray2'],
                                 line=dict(color='red'),
                                 name=vis_params['label2']))

        # Plot the 3rd option
        fig.add_trace(go.Scatter(x=vis_params['xarray'],
                                 y=vis_params['yarray3'],
                                 line=dict(color='green'),
                                 name=vis_params['label3']))

        # 4th option only used in Rho graphs
        if vis_params['label4'] is not None:
            fig.add_trace(go.Scatter(x=vis_params['xarray'],
                                     y=vis_params['yarray4'],
                                     line=dict(color='orange'),
                                     name=vis_params['label4']))
            rho_graph=True
        else:
            rho_graph=False

        xmin, xmax, ymin, ymax = cls._graph_range_2d(
            vis_params=vis_params, rho_graph=rho_graph)

        fig.update_layout(
            title={'text': vis_params['title'],
                   'y':0.95,
                   'x':0.5,
                   'xanchor':'center',
                   'yanchor':'top',
                   'font':dict(size=20,
                               color="#f2f5fa")},
            xaxis_title={'text': vis_params['xlabel'],
                         'font':dict(size=15,
                                     color="#f2f5fa")},
            yaxis_title={'text': vis_params['ylabel'],
                         'font':dict(size=15,
                                     color="#f2f5fa")},
            font={'color': '#f2f5fa'},
            paper_bgcolor='black',
            plot_bgcolor='black',
            legend=dict(
                x=0.05,
                y=0.95,
                traceorder="normal",
                bgcolor='rgba(0, 0, 0, 0)',
                font=dict(
                    family="sans-serif",
                    size=12,
                    color="#f2f5fa"
                ),
            ),
        )

        if params['web_graph'] is False:
            fig.update_layout(
                autosize=False,
                width=800,
                height=600
                )

        fig.update_xaxes(showline=True,
                         linewidth=2,
                         linecolor='#2a3f5f',
                         mirror=True,
                         range = [xmin, xmax],
                         gridwidth=1,
                         gridcolor='#2a3f5f',
                         zeroline=False)

        fig.update_yaxes(showline=True,
                         linewidth=2,
                         linecolor='#2a3f5f',
                         mirror=True,
                         range = [ymin, ymax],
                         gridwidth=1,
                         gridcolor='#2a3f5f',
                         zeroline=False)

        # If running in an iPython notebook the chart will display
        # in line
        if params['notebook']:
            # If output is sent to Dash
            if params['web_graph']:
                return fig

            fig.show()
            return

        # Otherwise create an HTML file that opens in a new window
        plot(fig, auto_open=True)
        return


    @staticmethod
    def _graph_range_2d(vis_params, rho_graph):
        """
        Set 2D graph ranges

        Parameters
        ----------
        vis_params : Dict
            Dictionary of parameters.

        Returns
        -------
        xmin : Float
            x-axis minimum.
        xmax : Float
            x-axis maximum.
        ymin : Float
            y-axis minimum.
        ymax : Float
            y-axis maximum.

        """

        ranges = {}
        ranges['min_x'] = vis_params['xarray'].min()
        ranges['max_x'] = vis_params['xarray'].max()
        ranges['min_y1'] = vis_params['yarray1'].min()
        ranges['max_y1'] = vis_params['yarray1'].max()
        ranges['min_y2'] = vis_params['yarray2'].min()
        ranges['max_y2'] = vis_params['yarray2'].max()
        ranges['min_y3'] = vis_params['yarray3'].min()
        ranges['max_y3'] = vis_params['yarray3'].max()

        if rho_graph:
            ranges['min_y4'] = vis_params['yarray4'].min()
            ranges['max_y4'] = vis_params['yarray4'].max()
        else:
            ranges['min_y4'] = ranges['min_y1']
            ranges['max_y4'] = ranges['max_y1']

        x_scale_shift = (ranges['max_x'] - ranges['min_x']) * 0.05
        xmin = ranges['min_x'] - x_scale_shift
        xmax = ranges['max_x'] + x_scale_shift
        y_scale_shift = (
            (max(ranges['max_y1'],
                 ranges['max_y2'],
                 ranges['max_y3'],
                 ranges['max_y4'])
            - min(ranges['min_y1'],
                  ranges['min_y2'],
                  ranges['min_y3'],
                  ranges['min_y4']))
            * 0.05)
        ymin = min(ranges['min_y1'],
                   ranges['min_y2'],
                   ranges['min_y3'],
                   ranges['min_y4']) - y_scale_shift
        ymax = max(ranges['max_y1'],
                   ranges['max_y2'],
                   ranges['max_y3'],
                   ranges['max_y4']) + y_scale_shift

        return xmin, xmax, ymin, ymax



"""
Option Pricing and Greeks formulas

"""
import numpy as np
from optionvisualizer.utilities import Utils
# pylint: disable=invalid-name

class Option():
    """
    Calculate Black Scholes Option Price and Greeks

    """

    @staticmethod
    def price(opt_params, params):
        """
        Black Scholes Option Price

        Parameters
        ----------
        opt_params : Dict
            S : Float
                Underlying Stock Price. The default is 100.
            K : Float
                Strike Price. The default is 100.
            T : Float
                Time to Maturity. The default is 0.25 (3 months).
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'
        params : Dict
            Dictionary of key parameters; used for refreshing distribution.

        Returns
        -------
        Float
            Black Scholes Option Price.

        """

        # Update distribution parameters
        params = Utils.refresh_dist_params(
            opt_params=opt_params, params=params)

        if opt_params['option'] == "call":
            opt_price = (
                (opt_params['S'] * params['carry'] * params['Nd1'])
                - (opt_params['K'] * np.exp(-opt_params['r'] * opt_params['T'])
                   * params['Nd2']))
        elif opt_params['option'] == "put":
            opt_price = (
                (opt_params['K'] * np.exp(-opt_params['r'] * opt_params['T'])
                 * params['minusNd2'])
                - (opt_params['S'] * params['carry'] * params['minusNd1']))

        else:
            print("Please supply an option type, 'put' or 'call'")

        np.nan_to_num(opt_price, copy=False)

        return opt_price


    @staticmethod
    def delta(opt_params, params):
        """
        Sensitivity of the option price to changes in asset price

        Parameters
        ----------
        opt_params : Dict
            S : Float
                Underlying Stock Price. The default is 100.
            K : Float
                Strike Price. The default is 100.
            T : Float
                Time to Maturity. The default is 0.25 (3 months).
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'
        params : Dict
            Dictionary of key parameters; used for refreshing distribution.

        Returns
        -------
        Float
            Option Delta.

        """

        # Update distribution parameters
        params = Utils.refresh_dist_params(
            opt_params=opt_params, params=params)

        if opt_params['option'] == 'call':
            opt_delta = params['carry'] * params['Nd1']
        if opt_params['option'] == 'put':
            opt_delta = params['carry'] * (params['Nd1'] - 1)

        return opt_delta


    @staticmethod
    def theta(opt_params, params):
        """
        Sensitivity of the option price to changes in time to maturity

        Parameters
        ----------
        opt_params : Dict
            S : Float
                Underlying Stock Price. The default is 100.
            K : Float
                Strike Price. The default is 100.
            T : Float
                Time to Maturity. The default is 0.25 (3 months).
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'
        params : Dict
            Dictionary of key parameters; used for refreshing distribution.

        Returns
        -------
        Float
            Option Theta.

        """

        # Update distribution parameters
        params = Utils.refresh_dist_params(
            opt_params=opt_params, params=params)

        if opt_params['option'] == 'call':
            opt_theta = (
                ((-opt_params['S']
                  * params['carry']
                  * params['nd1']
                  * opt_params['sigma'])
                 / (2 * np.sqrt(opt_params['T']))
                - (params['b'] - opt_params['r'])
                * (opt_params['S']
                   * params['carry']
                   * params['Nd1'])
                - (opt_params['r'] * opt_params['K'])
                * np.exp(-opt_params['r'] * opt_params['T'])
                * params['Nd2'])
                / 100)

        if opt_params['option'] == 'put':
            opt_theta = (
                ((-opt_params['S']
                  * params['carry']
                  * params['nd1']
                  * opt_params['sigma'] )
                 / (2 * np.sqrt(opt_params['T']))
                + (params['b'] - opt_params['r'])
                * (opt_params['S']
                   * params['carry']
                   * params['minusNd1'])
                + (opt_params['r'] * opt_params['K'])
                * np.exp(-opt_params['r'] * opt_params['T'])
                * params['minusNd2'])
                / 100)

        return opt_theta


    @staticmethod
    def gamma(opt_params, params):
        """
        Sensitivity of delta to changes in the underlying asset price

        Parameters
        ----------
        opt_params : Dict
            S : Float
                Underlying Stock Price. The default is 100.
            K : Float
                Strike Price. The default is 100.
            T : Float
                Time to Maturity. The default is 0.25 (3 months).
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'
        params : Dict
            Dictionary of key parameters; used for refreshing distribution.

        Returns
        -------
        Float
            Option Gamma.

        """

        # Update distribution parameters
        params = Utils.refresh_dist_params(
            opt_params=opt_params, params=params)

        opt_gamma = ((params['nd1'] * params['carry'])
                     / (opt_params['S'] * opt_params['sigma']
                        * np.sqrt(opt_params['T'])))

        return opt_gamma


    @staticmethod
    def vega(opt_params, params):
        """
        Sensitivity of the option price to changes in volatility

        Parameters
        ----------
        opt_params : Dict
            S : Float
                Underlying Stock Price. The default is 100.
            K : Float
                Strike Price. The default is 100.
            T : Float
                Time to Maturity. The default is 0.25 (3 months).
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'
        params : Dict
            Dictionary of key parameters; used for refreshing distribution.

        Returns
        -------
        Float
            Option Vega.

        """

        # Update distribution parameters
        params = Utils.refresh_dist_params(
            opt_params=opt_params, params=params)

        opt_vega = ((opt_params['S'] * params['carry']
                     * params['nd1'] * np.sqrt(opt_params['T'])) / 100)

        return opt_vega


    @staticmethod
    def rho(opt_params, params):
        """
        Sensitivity of the option price to changes in the risk free rate

        Parameters
        ----------
        opt_params : Dict
            S : Float
                Underlying Stock Price. The default is 100.
            K : Float
                Strike Price. The default is 100.
            T : Float
                Time to Maturity. The default is 0.25 (3 months).
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'
        params : Dict
            Dictionary of key parameters; used for refreshing distribution.

        Returns
        -------
        Float
            Option Rho.

        """

        # Update distribution parameters
        params = Utils.refresh_dist_params(
            opt_params=opt_params, params=params)

        if opt_params['option'] == 'call':
            opt_rho = (
                (opt_params['T'] * opt_params['K']
                 * np.exp(-opt_params['r'] * opt_params['T']) * params['Nd2'])
                / 10000)
        if opt_params['option'] == 'put':
            opt_rho = (
                (-opt_params['T'] * opt_params['K']
                 * np.exp(-opt_params['r'] * opt_params['T'])
                 * params['minusNd2'])
                / 10000)

        return opt_rho


    @staticmethod
    def vanna(opt_params, params):
        """
        DdeltaDvol, DvegaDspot
        Sensitivity of delta to changes in volatility
        Sensitivity of vega to changes in the asset price

        Parameters
        ----------
        opt_params : Dict
            S : Float
                Underlying Stock Price. The default is 100.
            K : Float
                Strike Price. The default is 100.
            T : Float
                Time to Maturity. The default is 0.25 (3 months).
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'
        params : Dict
            Dictionary of key parameters; used for refreshing distribution.

        Returns
        -------
        Float
            Option Vanna.

        """

        # Update distribution parameters
        params = Utils.refresh_dist_params(
            opt_params=opt_params, params=params)

        opt_vanna = (
            (((-params['carry'] * params['d2'])
              / opt_params['sigma']) * params['nd1']) / 100)

        return opt_vanna


    @classmethod
    def vomma(cls, opt_params, params):
        """
        DvegaDvol, Vega Convexity, Volga, Vol Gamma
        Sensitivity of vega to changes in volatility

        Parameters
        ----------
        opt_params : Dict
            S : Float
                Underlying Stock Price. The default is 100.
            K : Float
                Strike Price. The default is 100.
            T : Float
                Time to Maturity. The default is 0.25 (3 months).
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'
        params : Dict
            Dictionary of key parameters; used for refreshing distribution.

        Returns
        -------
        Float
            Option Vomma.

        """

        # Update distribution parameters
        params = Utils.refresh_dist_params(
            opt_params=opt_params, params=params)

        opt_vomma = (
            (cls.vega(opt_params, params) * (
                (params['d1'] * params['d2']) / (opt_params['sigma']))) / 100)

        return opt_vomma


    @staticmethod
    def charm(opt_params, params):
        """
        DdeltaDtime, Delta Bleed
        Sensitivity of delta to changes in time to maturity

        Parameters
        ----------
        opt_params : Dict
            S : Float
                Underlying Stock Price. The default is 100.
            K : Float
                Strike Price. The default is 100.
            T : Float
                Time to Maturity. The default is 0.25 (3 months).
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'
        params : Dict
            Dictionary of key parameters; used for refreshing distribution.

        Returns
        -------
        Float
            Option Charm.

        """

        # Update distribution parameters
        params = Utils.refresh_dist_params(
            opt_params=opt_params, params=params)

        if opt_params['option'] == 'call':
            opt_charm = (
                (-params['carry'] * ((params['nd1'] * (
                    (params['b'] / (opt_params['sigma']
                                    * np.sqrt(opt_params['T'])))
                    - (params['d2'] / (2 * opt_params['T']))))
                    + ((params['b'] - opt_params['r']) * params['Nd1'])))
                / 100)
        if opt_params['option'] == 'put':
            opt_charm = (
                (-params['carry'] * (
                    (params['nd1'] * (
                        (params['b']
                         / (opt_params['sigma'] * np.sqrt(opt_params['T'])))
                        - (params['d2'] / (2 * opt_params['T']))))
                    - ((params['b'] - opt_params['r']) * params['minusNd1'])))
                / 100)

        return opt_charm


    @classmethod
    def zomma(cls, opt_params, params):
        """
        DgammaDvol
        Sensitivity of gamma to changes in volatility

        Parameters
        ----------
        opt_params : Dict
            S : Float
                Underlying Stock Price. The default is 100.
            K : Float
                Strike Price. The default is 100.
            T : Float
                Time to Maturity. The default is 0.25 (3 months).
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'
        params : Dict
            Dictionary of key parameters; used for refreshing distribution.

        Returns
        -------
        Float
            Option Zomma.

        """

        # Update distribution parameters
        params = Utils.refresh_dist_params(
            opt_params=opt_params, params=params)

        opt_zomma = (
            (cls.gamma(opt_params, params) * (
                (params['d1'] * params['d2'] - 1)
                / opt_params['sigma'])) / 100)

        return opt_zomma


    @classmethod
    def speed(cls, opt_params, params):
        """
        DgammaDspot
        Sensitivity of gamma to changes in asset price
        3rd derivative of option price with respect to spot

        Parameters
        ----------
        opt_params : Dict
            S : Float
                Underlying Stock Price. The default is 100.
            K : Float
                Strike Price. The default is 100.
            T : Float
                Time to Maturity. The default is 0.25 (3 months).
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'
        params : Dict
            Dictionary of key parameters; used for refreshing distribution.

        Returns
        -------
        Float
            Option Speed.

        """

        # Update distribution parameters
        params = Utils.refresh_dist_params(
            opt_params=opt_params, params=params)

        opt_speed = -(
            cls.gamma(opt_params, params) * (1 + (
                params['d1'] / (opt_params['sigma']
                                * np.sqrt(opt_params['T']))))
            / opt_params['S'])

        return opt_speed


    @classmethod
    def color(cls, opt_params, params):
        """
        DgammaDtime, Gamma Bleed, Gamma Theta
        Sensitivity of gamma to changes in time to maturity

        Parameters
        ----------
        opt_params : Dict
            S : Float
                Underlying Stock Price. The default is 100.
            K : Float
                Strike Price. The default is 100.
            T : Float
                Time to Maturity. The default is 0.25 (3 months).
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'
        params : Dict
            Dictionary of key parameters; used for refreshing distribution.

        Returns
        -------
        Float
            Option Color.

        """

        # Update distribution parameters
        params = Utils.refresh_dist_params(
            opt_params=opt_params, params=params)

        opt_color = (
            (cls.gamma(opt_params, params) * (
                (opt_params['r'] - params['b']) + (
                    (params['b'] * params['d1'])
                    / (opt_params['sigma'] * np.sqrt(opt_params['T'])))
                + ((1 - params['d1'] * params['d2']) / (2 * opt_params['T']))))
            / 100)

        return opt_color


    @classmethod
    def ultima(cls, opt_params, params):
        """
        DvommaDvol
        Sensitivity of vomma to changes in volatility
        3rd derivative of option price wrt volatility

        Parameters
        ----------
        opt_params : Dict
            S : Float
                Underlying Stock Price. The default is 100.
            K : Float
                Strike Price. The default is 100.
            T : Float
                Time to Maturity. The default is 0.25 (3 months).
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'
        params : Dict
            Dictionary of key parameters; used for refreshing distribution.

        Returns
        -------
        Float
            Option Ultima.

        """

        # Update distribution parameters
        params = Utils.refresh_dist_params(
            opt_params=opt_params, params=params)

        opt_ultima = (
            (cls.vomma(opt_params, params) * (
                (1 / opt_params['sigma']) * (params['d1'] * params['d2']
                                         - (params['d1'] / params['d2'])
                                         - (params['d2'] / params['d1']) - 1)))
            / 100)

        return opt_ultima


    @classmethod
    def vega_bleed(cls, opt_params, params):
        """
        DvegaDtime
        Sensitivity of vega to changes in time to maturity.

        Parameters
        ----------
        opt_params : Dict
            S : Float
                Underlying Stock Price. The default is 100.
            K : Float
                Strike Price. The default is 100.
            T : Float
                Time to Maturity. The default is 0.25 (3 months).
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'
        params : Dict
            Dictionary of key parameters; used for refreshing distribution.

        Returns
        -------
        Float
            Option Vega Bleed.

        """

        # Update distribution parameters
        params = Utils.refresh_dist_params(
            opt_params=opt_params, params=params)

        opt_vega_bleed = (
            (cls.vega(opt_params, params)
             * (opt_params['r']
                - params['b']
                + ((params['b'] * params['d1'])
                   / (opt_params['sigma'] * np.sqrt(opt_params['T'])))
                - ((1 + (params['d1'] * params['d2']) )
                   / (2 * opt_params['T']))))
            / 100)

        return opt_vega_bleed


    @classmethod
    def return_options(cls, opt_dict, params):

        """
        Calculate option prices to be used in payoff diagrams.

        Parameters
        ----------
        opt_dict : Dict
            Dictionary of option pricing parameters
        params : Dict
            Dictionary of key parameters

        Returns
        -------
        From 1 to 4 sets of option values:
            Cx_0: Current option price; Float.
            Cx: Terminal Option payoff, varying by strike; Array
            Cx_G: Current option value, varying by strike; Array

        """

        # Dictionary to store option legs
        option_legs = {}

        # create array of 1000 equally spaced points between 75% of
        # initial underlying price and 125%
        option_legs['SA'] = np.linspace(
            0.75 * opt_dict['S'], 1.25 * opt_dict['S'], 1000)

        opt_params = {
            'S':opt_dict['S'],
            'K':opt_dict['K1'],
            'T':opt_dict['T1'],
            'r':opt_dict['r'],
            'q':opt_dict['q'],
            'sigma':opt_dict['sigma'],
            'option':opt_dict['option1'],
            }

        # Calculate the current price of option 1
        option_legs['C1_0'] = cls.price(opt_params=opt_params, params=params)

        # Calculate the prices at maturity for the range of strikes
        # in SA of option 1
        change_params = {'S':option_legs['SA'], 'T':0}
        opt_params.update(change_params)

        option_legs['C1'] = cls.price(opt_params=opt_params, params=params)

        # Calculate the current prices for the range of strikes
        # in SA of option 1
        change_params = {'T':opt_dict['T1']}
        opt_params.update(change_params)

        option_legs['C1_G'] = cls.price(opt_params=opt_params, params=params)

        if opt_dict['legs'] > 1:
            # Calculate the current price of option 2
            change_params = {'S':opt_dict['S'],
                             'K':opt_dict['K2'],
                             'T':opt_dict['T2'],
                             'option':opt_dict['option2']}
            opt_params.update(change_params)

            option_legs['C2_0'] = cls.price(
                opt_params=opt_params, params=params)

            # Calculate the prices at maturity for the range of strikes
            # in SA of option 2
            change_params = {'S':option_legs['SA'], 'T':0}
            opt_params.update(change_params)

            option_legs['C2'] = cls.price(opt_params=opt_params, params=params)

            # Calculate the current prices for the range of strikes
            # in SA of option 2
            change_params = {'T':opt_dict['T2']}
            opt_params.update(change_params)

            option_legs['C2_G'] = cls.price(
                opt_params=opt_params, params=params)

        if opt_dict['legs'] > 2:
            # Calculate the current price of option 3
            change_params = {'S':opt_dict['S'],
                             'K':opt_dict['K3'],
                             'T':opt_dict['T3'],
                             'option':opt_dict['option3']}
            opt_params.update(change_params)

            option_legs['C3_0'] = cls.price(
                opt_params=opt_params, params=params)

            # Calculate the prices at maturity for the range of strikes
            # in SA of option 3
            change_params = {'S':option_legs['SA'], 'T':0}
            opt_params.update(change_params)

            option_legs['C3'] = cls.price(opt_params=opt_params, params=params)

            # Calculate the current prices for the range of strikes
            # in SA of option 3
            change_params = {'T':opt_dict['T3']}
            opt_params.update(change_params)

            option_legs['C3_G'] = cls.price(
                opt_params=opt_params, params=params)

        if opt_dict['legs'] > 3:
            # Calculate the current price of option 4
            change_params = {'S':opt_dict['S'],
                             'K':opt_dict['K4'],
                             'T':opt_dict['T4'],
                             'option':opt_dict['option4']}
            opt_params.update(change_params)

            option_legs['C4_0'] = cls.price(
                opt_params=opt_params, params=params)

            # Calculate the prices at maturity for the range of strikes
            # in SA of option 4
            change_params = {'S':option_legs['SA'], 'T':0}
            opt_params.update(change_params)

            option_legs['C4'] = cls.price(
                opt_params=opt_params, params=params)

            # Calculate the current prices for the range of strikes
            # in SA of option 4
            change_params = {'T':opt_dict['T4']}
            opt_params.update(change_params)

            option_legs['C4_G'] = cls.price(
                opt_params=opt_params, params=params)

        return option_legs



"""
Key parameters for visualizer

"""

# pylint: disable=invalid-name

# Dictionary of default parameters
vis_params_dict = {
    'S':100,
    'K':100,
    'K1':95,
    'K2':105,
    'K3':105,
    'K4':105,
    'G1':90,
    'G2':100,
    'G3':110,
    'H':105,
    'R':0,
    'T':0.25,
    'T1':0.25,
    'T2':0.25,
    'T3':0.25,
    'T4':0.25,
    'r':0.005,
    'b':0.005,
    'q':0,
    'sigma':0.2,
    'eta':1,
    'phi':1,
    'C4':None,
    'label4':None,
    'barrier_direction':'down',
    'knock':'in',
    'option':'call',
    'option1':'call',
    'option2':'call',
    'option3':'call',
    'option4':'call',
    'direction':'long',
    'value':False,
    'ratio':2,
    'refresh':'Std',
    'combo_payoff':'straddle',
    'payoff_type':'straddle',
    'price_shift':0.25,
    'price_shift_type':'avg',
    'vol_shift':0.001,
    'ttm_shift':1/365,
    'rate_shift':0.0001,
    'greek':'delta',
    'num_sens':False,
    'interactive':False,
    'notebook':True,
    'web_graph':False,
    'colorscheme':'jet',
    'colorintensity':1,
    'size3d':(15, 12),
    'size2d':(8, 6),
    'graphtype':'2D',
    'y_plot':'delta',
    'x_plot':'price',
    'time_shift':0.25,
    'cash':False,
    'axis':'price',
    'spacegrain':100,
    'azim':-60,
    'elev':30,
    'risk':True,
    'gif':False,
    'gif_folder':None,
    'gif_filename':None,
    'gif_folder_2D':'images/greeks2D',
    'gif_filename_2D':'greek2D',
    'gif_folder_3D':'images/greeks3D',
    'gif_filename_3D':'greek3D',
    'gif_frame_update':2,
    'gif_min_dist':9.0,
    'gif_max_dist':10.0,
    'gif_min_elev':10.0,
    'gif_max_elev':60.0,
    'gif_dpi':50,
    'gif_ms':100,
    'gif_start_azim':0,
    'gif_end_azim':360,
    'mpl_style':'seaborn-darkgrid',
    'steps':40,
    'titlename':None,
    'title_font_scale':None,

     # List of default parameters used when refreshing
     'params_list':[
         'S', 'K', 'K1', 'K2', 'K3', 'K4', 'G1', 'G2', 'G3', 'H', 'R',
         'T', 'T1', 'T2', 'T3', 'T4', 'r', 'b', 'q', 'sigma', 'eta',
         'phi', 'barrier_direction', 'knock', 'option', 'option1',
         'option2', 'option3', 'option4', 'direction', 'value',
         'ratio', 'refresh', 'price_shift', 'price_shift_type',
         'vol_shift','ttm_shift', 'rate_shift', 'greek', 'num_sens',
         'interactive', 'notebook', 'colorscheme', 'colorintensity',
         'size3d', 'size2d', 'graphtype', 'cash', 'axis', 'spacegrain',
         'azim', 'elev', 'risk', 'mpl_style'
         ],

     # List of Greeks where call and put values are the same
     'equal_greeks':[
         'gamma',
         'vega',
         'vomma',
         'vanna',
         'zomma',
         'speed',
         'color',
         'ultima',
         'vega bleed'
         ],

     # Payoffs requiring changes to default parameters
     'mod_payoffs':[
         'call',
         'put',
         'collar',
         'straddle',
         'butterfly',
         'christmas tree',
         'condor',
         'iron butterfly',
         'iron condor'
         ],

     # Those parameters that need changing
     'mod_params':[
         'S',
         'K',
         'K1',
         'K2',
         'K3',
         'K4',
         'T',
         'T1',
         'T2',
         'T3',
         'T4'
         ],

     # Combo parameter values differing from standard defaults
     'combo_dict':{
         'call':{
             'S':100,
             'K':100,
             'K1':100,
             'T1':0.25
             },
         'put':{
             'S':100,
             'K':100,
             'K1':100,
             'T1':0.25
             },
         'collar':{
             'S':100,
             'K':100,
             'K1':98,
             'K2':102,
             'T1':0.25,
             'T2':0.25
             },
         'straddle':{
             'S':100,
             'K':100,
             'K1':100,
             'K2':100,
             'T1':0.25,
             'T2':0.25
             },
         'butterfly':{
             'S':100,
             'K':100,
             'K1':95,
             'K2':100,
             'K3':105,
             'T1':0.25,
             'T2':0.25,
             'T3':0.25
             },
         'christmas tree':{
             'S':100,
             'K':100,
             'K1':95,
             'K2':100,
             'K3':105,
             'T1':0.25,
             'T2':0.25,
             'T3':0.25
             },
         'condor':{
             'S':100,
             'K':100,
             'K1':90,
             'K2':95,
             'K3':100,
             'K4':105,
             'T1':0.25,
             'T2':0.25,
             'T3':0.25,
             'T4':0.25
             },
         'iron butterfly':{
             'S':100,
             'K':100,
             'K1':95,
             'K2':100,
             'K3':100,
             'K4':105,
             'T1':0.25,
             'T2':0.25,
             'T3':0.25,
             'T4':0.25
             },
         'iron condor':{
             'S':100,
             'K':100,
             'K1':90,
             'K2':95,
             'K3':100,
             'K4':105,
             'T1':0.25,
             'T2':0.25,
             'T3':0.25,
             'T4':0.25
             }
         },

    'combo_parameters':{
        'call':[
            'S',
            'K',
            'T',
            'r',
            'q',
            'sigma',
            'direction',
            'value',
            'mpl_style',
            'size2d'
            ],
        'put':[
            'S',
            'K',
            'T',
            'r',
            'q',
            'sigma',
            'direction',
            'value',
            'mpl_style',
            'size2d'
            ],
        'stock':[
            'S',
            'direction',
            'mpl_style',
            'size2d'
            ],
        'forward':[
            'S',
            'T',
            'r',
            'q',
            'sigma',
            'direction',
            'cash',
            'mpl_style',
            'size2d'
            ],
        'collar':[
            'S',
            'K1',
            'K2',
            'T',
            'r',
            'q',
            'sigma',
            'direction',
            'value',
            'mpl_style',
            'size2d'
            ],
        'spread':[
            'S',
            'K1',
            'K2',
            'T',
            'r',
            'q',
            'sigma',
            'option',
            'direction',
            'value',
            'mpl_style',
            'size2d'
            ],
        'backspread':[
            'S',
            'K1',
            'K2',
            'T',
            'r',
            'q',
            'sigma',
            'option',
            'ratio',
            'value',
            'mpl_style',
            'size2d'
            ],
        'ratio vertical spread':[
            'S',
            'K1',
            'K2',
            'T',
            'r',
            'q',
            'sigma',
            'option',
            'ratio',
            'value',
            'mpl_style',
            'size2d'
            ],
        'straddle':[
            'S',
            'K',
            'T',
            'r',
            'q',
            'sigma',
            'direction',
            'value',
            'mpl_style',
            'size2d'
            ],
        'strangle':[
            'S',
            'K1',
            'K2',
            'T',
            'r',
            'q',
            'sigma',
            'direction',
            'value',
            'mpl_style',
            'size2d'
            ],
        'butterfly':[
            'S',
            'K1',
            'K2',
            'K3',
            'T',
            'r',
            'q',
            'sigma',
            'option',
            'direction',
            'value',
            'mpl_style',
            'size2d'
            ],
        'christmas tree':[
            'S',
            'K1',
            'K2',
            'K3',
            'T',
            'r',
            'q',
            'sigma',
            'option',
            'direction',
            'value',
            'mpl_style',
            'size2d'
            ],
        'condor':[
            'S',
            'K1',
            'K2',
            'K3',
            'K4',
            'T',
            'r',
            'q',
            'sigma',
            'option',
            'direction',
            'value',
            'mpl_style',
            'size2d'
            ],
        'iron butterfly':[
            'S',
            'K1',
            'K2',
            'K3',
            'K4',
            'T',
            'r',
            'q',
            'sigma',
            'direction',
            'value',
            'mpl_style',
            'size2d'
            ],
        'iron condor':[
            'S',
            'K1',
            'K2',
            'K3',
            'K4',
            'T',
            'r',
            'q',
            'sigma',
            'direction',
            'value',
            'mpl_style',
            'size2d'
            ],
        },

    'combo_name_dict':{
        'call':'call',
        'put':'put',
        'stock':'stock',
        'forward':'forward',
        'collar':'collar',
        'spread':'spread',
        'backspread':'backspread',
        'ratio vertical spread':'ratio_vertical_spread',
        'straddle':'straddle',
        'strangle':'strangle',
        'butterfly':'butterfly',
        'christmas tree':'christmas_tree',
        'condor':'condor',
        'iron butterfly':'iron_butterfly',
        'iron condor':'iron_condor',
        },

    'combo_simple_dict':{
        'call':'call',
        'put':'put',
        'stock':'stock',
        'forward':'forward',
        'collar':'collar',
        'spread':'spread',
        'backspread':'backspread',
        'ratio vertical spread':'ratio_vertical_spread',
        'straddle':'straddle',
        'strangle':'strangle',
        },

    'combo_multi_dict':{
        'butterfly':'butterfly',
        'christmas tree':'christmas_tree',
        'condor':'condor',
        'iron butterfly':'iron_butterfly',
        'iron condor':'iron_condor',
        },

     # Dictionary mapping function parameters to x axis labels
     # for 2D graphs
     'x_name_dict':{
         'price':'SA',
         'strike':'SA',
         'vol':'sigmaA',
         'time':'TA'
         },

     # Dictionary mapping scaling parameters to x axis labels
     # for 2D graphs
     'x_scale_dict':{
         'price':1,
         'strike':1,
         'vol':100,
         'time':365
         },

     # Dictionary mapping function parameters to y axis labels
     # for 2D graphs
     'y_name_dict':{
         'value':'price',
         'delta':'delta',
         'gamma':'gamma',
         'vega':'vega',
         'theta':'theta'
         },

     # Dictionary mapping function parameters to axis labels
     # for 3D graphs
     'label_dict':{
         'price':'Underlying Price',
         'value':'Theoretical Value',
         'vol':'Volatility %',
         'time':'Time to Expiration (Days)',
         'delta':'Delta',
         'gamma':'Gamma',
         'vega':'Vega',
         'theta':'Theta',
         'rho':'Rho',
         'strike':'Strike Price'
         },

     # Ranges of Underlying price and Time to Expiry for 3D
     # greeks graphs
     '3D_chart_ranges':{
         'price':{
             'SA_lower':0.8,
             'SA_upper':1.2,
             'TA_lower':0.01,
             'TA_upper':1
             },
         'delta':{
             'SA_lower':0.25,
             'SA_upper':1.75,
             'TA_lower':0.01,
             'TA_upper':2
             },
         'gamma':{
             'SA_lower':0.8,
             'SA_upper':1.2,
             'TA_lower':0.01,
             'TA_upper':5
             },
         'vega':{
             'SA_lower':0.5,
             'SA_upper':1.5,
             'TA_lower':0.01,
             'TA_upper':1
             },
         'theta':{
             'SA_lower':0.8,
             'SA_upper':1.2,
             'TA_lower':0.01,
             'TA_upper':1
             },
         'rho':{
             'SA_lower':0.8,
             'SA_upper':1.2,
             'TA_lower':0.01,
             'TA_upper':0.5
             },
         'vomma':{
             'SA_lower':0.5,
             'SA_upper':1.5,
             'TA_lower':0.01,
             'TA_upper':1
             },
         'vanna':{
             'SA_lower':0.5,
             'SA_upper':1.5,
             'TA_lower':0.01,
             'TA_upper':1
             },
         'zomma':{
             'SA_lower':0.8,
             'SA_upper':1.2,
             'TA_lower':0.01,
             'TA_upper':0.5
             },
         'speed':{
             'SA_lower':0.8,
             'SA_upper':1.2,
             'TA_lower':0.01,
             'TA_upper':0.5
             },
         'color':{
             'SA_lower':0.8,
             'SA_upper':1.2,
             'TA_lower':0.01,
             'TA_upper':0.5
             },
         'ultima':{
             'SA_lower':0.5,
             'SA_upper':1.5,
             'TA_lower':0.01,
             'TA_upper':1
             },
         'vega bleed':{
             'SA_lower':0.5,
             'SA_upper':1.5,
             'TA_lower':0.01,
             'TA_upper':1
             },
         'charm':{
             'SA_lower':0.8,
             'SA_upper':1.2,
             'TA_lower':0.01,
             'TA_upper':0.25
             }
         },

     # Greek names as function input and individual function
     # names
     'greek_dict':{
         'price':'price',
         'delta':'delta',
         'gamma':'gamma',
         'vega':'vega',
         'theta':'theta',
         'rho':'rho',
         'vomma':'vomma',
         'vanna':'vanna',
         'zomma':'zomma',
         'speed':'speed',
         'color':'color',
         'ultima':'ultima',
         'vega bleed':'vega_bleed',
         'charm':'charm'
         },

     # Parameters to overwrite mpl_style defaults
     'mpl_params':{
         'legend.fontsize': 'x-large',
         'legend.fancybox':False,
         'figure.dpi':72,
         'axes.labelsize': 'medium',
         'axes.titlesize':'large',
         'axes.spines.bottom':True,
         'axes.spines.left':True,
         'axes.spines.right':True,
         'axes.spines.top':True,
         'axes.edgecolor':'black',
         'axes.titlepad':20,
         'axes.autolimit_mode':'data',#'round_numbers',
         'axes.xmargin':0.05,
         'axes.ymargin':0.05,
         'axes.linewidth':2,
         'xtick.labelsize':'medium',
         'ytick.labelsize':'medium',
         'xtick.major.pad':10,
         'ytick.major.pad':10,
         'lines.linewidth':3.0,
         'lines.color':'black',
         'grid.color':'black',
         'grid.linestyle':':',
         'font.size':14
         },

     'mpl_3d_params':{
         'axes.facecolor':'w',
         'axes.labelcolor':'k',
         'axes.edgecolor':'w',
         'axes.titlepad':5,
         'lines.linewidth':0.5,
         'xtick.labelbottom':True,
         'ytick.labelleft':True
         }
     }



"""
Calculate Analytical or Numerical sensitivities

"""

from optionvisualizer.option_formulas import Option
# pylint: disable=invalid-name

class Sens():
    """
    Summary functions for calculating sensitivities

    """
    @classmethod
    def sensitivities_static(cls, params, **kwargs):
        """
        Sensitivities of the option.

        Parameters
        ----------
        S : Float
            Underlying Stock Price. The default is 100.
        K : Float
            Strike Price. The default is 100.
        T : Float
            Time to Maturity. The default is 0.25 (3 months).
        r : Float
            Interest Rate. The default is 0.05 (5%).
        q : Float
            Dividend Yield. The default is 0.
        sigma : Float
            Implied Volatility. The default is 0.2 (20%).
        option : Str
            Option type, Put or Call. The default is 'call'
        greek : Str
            Sensitivity to return. Select from 'delta', 'gamma', 'vega',
            'theta', 'rho', 'vomma', 'vanna', 'zomma', 'speed', 'color',
            'ultima', 'vega bleed', 'charm'. The default is 'delta'
        price_shift : Float
            The size of the price shift in decimal terms. The default
            is 0.25.
        vol_shift : Float
            The size of the volatility shift in decimal terms. The
            default is 0.001.
        ttm_shift : Float
            The size of the time to maturity shift in decimal terms. The
            default is 1/365.
        rate_shift : Float
            The size of the interest rate shift in decimal terms. The
            default is 0.0001.
        num_sens : Bool
            Whether to calculate numerical or analytical sensitivity.
            The default is False.

        Returns
        -------
        Float
            Option Sensitivity.

        """

        opt_params = {}
        op_keys = ['S','K', 'T', 'r', 'q', 'sigma', 'option']

        # Update params with the specified parameters
        for key, value in kwargs.items():

            if key not in ['params']:
                if key in op_keys:
                    # Add to the option parameter dictionary
                    opt_params[key] = value
                # Replace the default parameter with that provided
                else:
                    params[key] = value

        for parameter in op_keys:
            if parameter not in opt_params:
                opt_params[parameter] = params[parameter]

        if params['num_sens']:
            return cls.numerical_sensitivities(
                opt_params=opt_params, params=params)

        return cls.analytical_sensitivities(
            opt_params=opt_params, params=params)


    @staticmethod
    def analytical_sensitivities(opt_params, params):
        """
        Sensitivities of the option calculated analytically from closed
        form solutions.

        Parameters
        ----------
        S : Float
            Underlying Stock Price. The default is 100.
        K : Float
            Strike Price. The default is 100.
        T : Float
            Time to Maturity. The default is 0.25 (3 months).
        r : Float
            Interest Rate. The default is 0.05 (5%).
        q : Float
            Dividend Yield. The default is 0.
        sigma : Float
            Implied Volatility. The default is 0.2 (20%).
        option : Str
            Option type, Put or Call. The default is 'call'
        greek : Str
            Sensitivity to return. Select from 'delta', 'gamma', 'vega',
            'theta', 'rho', 'vomma', 'vanna', 'zomma', 'speed', 'color',
            'ultima', 'vega bleed', 'charm'. The default is 'delta'

        Returns
        -------
        Float
            Option Sensitivity.

        """

        for key, value in params['greek_dict'].items():
            if str(params['greek']) == key:
                return getattr(Option, value)(
                    opt_params=opt_params, params=params)

        return print("Please enter a valid Greek")


    @classmethod
    def numerical_sensitivities(cls, opt_params, params):
        """
        Sensitivities of the option calculated numerically using shifts
        in parameters.

        Parameters
        ----------
        S : Float
            Underlying Stock Price. The default is 100.
        K : Float
            Strike Price. The default is 100.
        T : Float
            Time to Maturity. The default is 0.25 (3 months).
        r : Float
            Interest Rate. The default is 0.05 (5%).
        q : Float
            Dividend Yield. The default is 0.
        sigma : Float
            Implied Volatility. The default is 0.2 (20%).
        option : Str
            Option type, Put or Call. The default is 'call'
        greek : Str
            Sensitivity to return. Select from 'delta', 'gamma', 'vega',
            'theta', 'rho', 'vomma', 'vanna', 'zomma', 'speed', 'color',
            'ultima', 'vega bleed', 'charm'. The default is 'delta'
        price_shift : Float
            The size of the price shift in decimal terms. The default
            is 0.25.
        vol_shift : Float
            The size of the volatility shift in decimal terms. The
            default is 0.001.
        ttm_shift : Float
            The size of the time to maturity shift in decimal terms. The
            default is 1/365.
        rate_shift : Float
            The size of the interest rate shift in decimal terms. The
            default is 0.0001.

        Returns
        -------
        Float
            Option Sensitivity.

        """

        for key, value in params['greek_dict'].items():
            if key == params['greek']:
                return getattr(cls, '_num_'+value)(
                    opt_params=opt_params, params=params)

        return print("Please enter a valid Greek")


    @classmethod
    def _num_price(cls, opt_params, params):
        option_names = []
        opt_dict = cls._option_prices(
            opt_params=opt_params, params=params, option_names=option_names)
        return opt_dict['opt']


    @classmethod
    def _num_delta(cls, opt_params, params):
        option_names = ['price_up', 'price_down']
        opt_dict = cls._option_prices(
            opt_params=opt_params, params=params, option_names=option_names)
        return ((opt_dict['opt_price_up'] - opt_dict['opt_price_down'])
                / (2 * params['price_shift']))


    @classmethod
    def _num_gamma(cls, opt_params, params):
        option_names = ['price_up', 'price_down']
        opt_dict = cls._option_prices(
            opt_params=opt_params, params=params, option_names=option_names)
        return ((opt_dict['opt_price_up']
                 - (2 * opt_dict['opt'])
                 + opt_dict['opt_price_down'])
                / (params['price_shift'] ** 2))


    @classmethod
    def _num_vega(cls, opt_params, params):
        option_names = ['vol_up', 'vol_down']
        opt_dict = cls._option_prices(
            opt_params=opt_params, params=params, option_names=option_names)
        return (((opt_dict['opt_vol_up'] - opt_dict['opt_vol_down'])
                 / (2 * params['vol_shift'])) / 100)


    @classmethod
    def _num_theta(cls, opt_params, params):
        option_names = ['ttm_down']
        opt_dict = cls._option_prices(
            opt_params=opt_params, params=params, option_names=option_names)
        return ((opt_dict['opt_ttm_down'] - opt_dict['opt'])
                / (params['ttm_shift'] * 100))


    @classmethod
    def _num_rho(cls, opt_params, params):
        option_names = ['rate_up', 'rate_down']
        opt_dict = cls._option_prices(
            opt_params=opt_params, params=params, option_names=option_names)
        return ((opt_dict['opt_rate_up'] - opt_dict['opt_rate_down'])
                / (2 * params['rate_shift'] * 10000))


    @classmethod
    def _num_vomma(cls, opt_params, params):
        option_names = ['vol_up', 'vol_down']
        opt_dict = cls._option_prices(
            opt_params=opt_params, params=params, option_names=option_names)
        return (((opt_dict['opt_vol_up']
                  - (2 * opt_dict['opt'])
                  + opt_dict['opt_vol_down'])
                 / (params['vol_shift'] ** 2)) / 10000)


    @classmethod
    def _num_vanna(cls, opt_params, params):
        option_names = ['price_up_vol_up', 'price_up_vol_down',
                        'price_down_vol_up', 'price_down_vol_down']
        opt_dict = cls._option_prices(
            opt_params=opt_params, params=params, option_names=option_names)
        return (((1 / (4 * params['price_shift'] * params['vol_shift']))
                 * (opt_dict['opt_price_up_vol_up']
                    - opt_dict['opt_price_up_vol_down']
                    - opt_dict['opt_price_down_vol_up']
                    + opt_dict['opt_price_down_vol_down'])) / 100)


    @classmethod
    def _num_charm(cls, opt_params, params):
        option_names = ['price_up', 'price_down',
                        'price_up_ttm_down', 'price_down_ttm_down']
        opt_dict = cls._option_prices(
            opt_params=opt_params, params=params, option_names=option_names)
        return ((((opt_dict['opt_price_up_ttm_down']
                   - opt_dict['opt_price_down_ttm_down'])
                  / (2 * params['price_shift']))
                     - ((opt_dict['opt_price_up']
                         - opt_dict['opt_price_down'])
                        / (2 * params['price_shift'])))
                    / (params['ttm_shift'] * 100))


    @classmethod
    def _num_zomma(cls, opt_params, params):
        option_names = ['vol_up', 'vol_down', 'price_up_vol_up',
                        'price_up_vol_down', 'price_down_vol_up',
                        'price_down_vol_down']
        opt_dict = cls._option_prices(
            opt_params=opt_params, params=params, option_names=option_names)
        return (((opt_dict['opt_price_up_vol_up']
                  - (2 * opt_dict['opt_vol_up'])
                  + opt_dict['opt_price_down_vol_up'])
                 - opt_dict['opt_price_up_vol_down']
                 + (2 * opt_dict['opt_vol_down'])
                 - opt_dict['opt_price_down_vol_down'])
                / (2 * params['vol_shift'] * (params['price_shift'] ** 2))
                / 100)


    @classmethod
    def _num_speed(cls, opt_params, params):
        option_names = ['price_up', 'price_down', 'double_price_up']
        opt_dict = cls._option_prices(
            opt_params=opt_params, params=params, option_names=option_names)
        return (1 / (params['price_shift'] ** 3)
                * (opt_dict['opt_double_price_up']
                   - (3 * opt_dict['opt_price_up'])
                   + 3 * opt_dict['opt']
                   - opt_dict['opt_price_down']))


    @classmethod
    def _num_color(cls, opt_params, params):
        option_names = ['price_up', 'price_down', 'price_up_ttm_down',
                        'price_down_ttm_down', 'ttm_down']
        opt_dict = cls._option_prices(
            opt_params=opt_params, params=params, option_names=option_names)
        return ((((opt_dict['opt_price_up_ttm_down']
                   - (2 * opt_dict['opt_ttm_down'])
                   + opt_dict['opt_price_down_ttm_down'])
                  / (params['price_shift'] ** 2))
                 - ((opt_dict['opt_price_up']
                     - (2 * opt_dict['opt'])
                     + opt_dict['opt_price_down'])
                    / (params['price_shift'] ** 2) ))
                / (params['ttm_shift'] * 100))


    @classmethod
    def _num_ultima(cls, opt_params, params):
        option_names = ['vol_up', 'vol_down', 'double_vol_up']
        opt_dict = cls._option_prices(
            opt_params=opt_params, params=params, option_names=option_names)
        return ((1 / (params['vol_shift'] ** 3)
                 * (opt_dict['opt_double_vol_up']
                    - (3 * opt_dict['opt_vol_up'])
                    + 3 * opt_dict['opt']
                    - opt_dict['opt_vol_down']))
                * (params['vol_shift'] ** 2))


    @classmethod
    def _num_vega_bleed(cls, opt_params, params):
        option_names = ['vol_up', 'vol_down', 'vol_up_ttm_down',
                        'vol_down_ttm_down']
        opt_dict = cls._option_prices(
            opt_params=opt_params, params=params, option_names=option_names)
        return ((((opt_dict['opt_vol_up_ttm_down']
                   - opt_dict['opt_vol_down_ttm_down'])
                  / (2 * params['vol_shift']))
                 - ((opt_dict['opt_vol_up']
                     - opt_dict['opt_vol_down'])
                    / (2 * params['vol_shift'])))
                / (params['ttm_shift'] * 10000))


    @classmethod
    def _option_prices(cls, opt_params, params, option_names):

        if params['greek'] in params['equal_greeks']:
            opt_params['option'] = 'call'

        shift_dict = cls._parameter_shifts(
            opt_params=opt_params, params=params)

        opt_dict = {}
        opt_dict['opt'] = Option.price(opt_params=opt_params, params=params)

        for option_name, option_params in shift_dict.items():
            if option_name in option_names:
                opt_dict['opt_'+option_name] = Option.price(
                opt_params=option_params, params=params)

        return opt_dict


    @staticmethod
    def _parameter_shifts(opt_params, params):

        shift_dict = {}
        shift_inputs = {'price':'S',
                        'vol':'sigma',
                        'ttm':'T',
                        'rate':'r'}

        for shift, shift_param in shift_inputs.items():
            shift_dict[shift+'_up'] = {
                **opt_params, shift_param:(
                    opt_params[shift_param] + params[shift+'_shift'])}
            shift_dict[shift+'_down'] = {
                **opt_params, shift_param:(
                    opt_params[shift_param] - params[shift+'_shift'])}

        shift_dict['price_up_vol_up'] ={
            **opt_params,
            'S':(opt_params['S'] + params['price_shift']),
            'sigma':(opt_params['sigma'] + params['vol_shift'])}

        shift_dict['price_down_vol_up'] ={
            **opt_params,
            'S':(opt_params['S'] - params['price_shift']),
            'sigma':(opt_params['sigma'] + params['vol_shift'])}

        shift_dict['price_up_vol_down'] ={
            **opt_params,
            'S':(opt_params['S'] + params['price_shift']),
            'sigma':(opt_params['sigma'] - params['vol_shift'])}

        shift_dict['price_down_vol_down'] ={
            **opt_params,
            'S':(opt_params['S'] - params['price_shift']),
            'sigma':(opt_params['sigma'] - params['vol_shift'])}

        shift_dict['price_up_ttm_down'] ={
            **opt_params,
            'S':(opt_params['S'] + params['price_shift']),
            'T':(opt_params['T'] - params['ttm_shift'])}

        shift_dict['price_down_ttm_down'] ={
            **opt_params,
            'S':(opt_params['S'] - params['price_shift']),
            'T':(opt_params['T'] - params['ttm_shift'])}

        shift_dict['vol_up_ttm_down'] ={
            **opt_params,
            'sigma':(opt_params['sigma'] + params['vol_shift']),
            'T':(opt_params['T'] - params['ttm_shift'])}

        shift_dict['vol_down_ttm_down'] ={
            **opt_params,
            'sigma':(opt_params['sigma'] - params['vol_shift']),
            'T':(opt_params['T'] - params['ttm_shift'])}

        shift_dict['double_price_up'] = {
            **opt_params,
            'S':(opt_params['S'] + 2 * params['price_shift'])}

        shift_dict['double_vol_up'] = {
            **opt_params,
            'sigma':(opt_params['sigma'] + 2 * params['vol_shift'])}

        return shift_dict


    @staticmethod
    def option_data_static(params, option_value, **kwargs):
        """
        Calculate Option prices or Greeks

        Parameters
        ----------
        option_value : str
            The value to return; price or specified greek
        params : Dict
            S : Float
                Underlying Stock Price. The default is 100.
            K : Float
                Strike Price. The default is 100.
            T : Float
                Time to Maturity. The default is 0.25 (3 months).
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'


        Returns
        -------
        Price / Greek.

        """
        opt_params ={}

        # Update params with the specified parameters
        for key, value in kwargs.items():
            if key not in ['params']:
                if key in ['S','K', 'T', 'r', 'q', 'sigma', 'option']:
                    # Add to the option parameter dictionary
                    opt_params[key] = value
                # Replace the default parameter with that provided
                else:
                    params[key] = value

        try:
            # Select the chosen option value from the available functions
            function = params['greek_dict'][option_value]
            return getattr(Option, function)(
                opt_params=opt_params, params=params)

        except KeyError:
            return print("Please enter a valid function from 'price', "\
                   "'delta', 'gamma', 'vega', 'theta', 'rho', 'vomma', "\
                       "'vanna', 'zomma', 'speed', 'color', 'ultima', "\
                           "'vega bleed', 'charm'")



"""
Create Animated Gifs

"""
import glob
import math
import os
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image
from optionvisualizer.greeks import Greeks
# pylint: disable=invalid-name

class Gif():
    """
    Create Animated Gifs

    """
    @classmethod
    def animated_2D_gif(cls, params):
        """
        Create an animated gif of the selected pair of parameters.

        Parameters
        ----------
        params : Dict
            gif_folder : Str
                The folder to save the files into. The default is
                'images/greeks'.
            gif_filename : Str
                The filename for the animated gif. The default is 'greek'.
            T : Float
                Time to Maturity. The default is 0.5 (6 months).
            steps : Int
                Number of images to combine. The default is 40.
            x_plot : Str
                     The x-axis variable ('price', 'strike' or 'vol').
                     The default is 'price'.
            y_plot : Str
                     The y-axis variable ('value', 'delta', 'gamma', 'vega'
                     or 'theta'). The default is 'delta.
            S : Float
                 Underlying Stock Price. The default is 100.
            G1 : Float
                 Strike Price of option 1. The default is 90.
            G2 : Float
                 Strike Price of option 2. The default is 100.
            G3 : Float
                 Strike Price of option 3. The default is 110.
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'
            direction : Str
                Whether the payoff is long or short. The default is 'long'.
            size2d : Tuple
                Figure size for matplotlib chart. The default is (6, 4).
            mpl_style : Str
                Matplotlib style template for 2D risk charts and payoffs.
                The default is 'seaborn-darkgrid'.
            num_sens : Bool
                Whether to calculate numerical or analytical sensitivity.
                The default is False.

        Returns
        -------
        Saves an animated gif.

        """
        params['gif']=True

        # Set up folders to save files
        params = cls._gif_defaults_setup(params=params)

        # split the countdown from T to maturity in steps equal steps
        time_steps = np.linspace(params['T'], 0.001, params['steps'])

        # create a plot for each time_step
        for counter, step in enumerate(time_steps):

            # create filenumber and filename
            filenumber = '{:03d}'.format(counter)
            filename = '{}, {}'.format(
                params['gif_filename'], filenumber)

            params['T'] = step
            params['T1'] = step
            params['T2'] = step
            params['T3'] = step

            # call the greeks_graphs_2d function to create graph
            fig, ax = Greeks.greeks_graphs_2D(
                params=params)

            # save the image as a file
            plt.savefig('{}/{}/img{}.png'.format(
                params['gif_folder'], params['gif_filename'], filename),
                dpi=50)

            # close the image object
            plt.close()

        cls._create_animation(params)


    @classmethod
    def animated_3D_gif(cls, params):
        """
        Create an animated gif of the selected greek 3D graph.

        Parameters
        ----------
        params : Dict
            S : Float
                 Underlying Stock Price. The default is 100.
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'
            direction : Str
                Whether the payoff is long or short. The default is 'long'.
            notebook : Bool
                Whether the function is being run in an IPython notebook and
                hence whether it should output in line or to an HTML file.
                The default is False.
            colorscheme : Str
                The matplotlib colormap or plotly colorscale to use. The
                default is 'jet' (which works in both).
            colorintensity : Float
                The alpha value indicating level of transparency /
                intensity. The default is 1.
            size3d : Tuple
                Figure size for matplotlib chart. The default is (15, 12).
            axis : Str
                Whether the x-axis is 'price' or 'vol'. The default
                is 'price'.
            spacegrain : Int
                Number of points in each axis linspace argument for 3D
                graphs. The default is 100.
            azim : Float
                L-R view angle for 3D graphs. The default is -50.
            elev : Float
                Elevation view angle for 3D graphs. The default is 20.
            greek : Str
                The sensitivity to be charted. Select from 'delta', 'gamma',
                'vega', 'theta', 'rho', 'vomma', 'vanna', 'zomma', 'speed',
                'color', 'ultima', 'vega_bleed', 'charm'. The default is
                'delta'
            num_sens : Bool
                Whether to calculate numerical or analytical sensitivity.
                The default is False.
            gif_folder : Str
                The folder to save the files into. The default is
                'images/greeks'.
            gif_filename : Str
                The filename for the animated gif. The default is 'greek'.
            gif_frame_update : Int
                The number of degrees of rotation between each frame used to
                construct the animated gif. The default is 2.
            gif_min_dist : Float
                The minimum zoom distance. The default is 9.0.
            gif_max_dist : Float
                The maximum zoom distance. The default is 10.0.
            gif_min_elev : Float
                The minimum elevation. The default is 10.0.
            gif_max_elev : Float
                The maximum elevation. The default is 60.0.
            gif_start_azim : Float
                The azimuth to start the gif from. The default is 0.
            gif_end_azim : Float
                The azimuth to end the gif on. The default is 360.
            gif_dpi : Int
                The image resolution to save. The default is 50 dpi.
            gif_ms : Int
                The time to spend on each frame in the gif. The default is
                100ms.

        Returns
        -------
        Saves an animated gif.

        """

        params = cls._gif_defaults_setup(params=params)

        fig, ax, params['titlename'], \
            params['title_font_scale'] = Greeks.greeks_graphs_3D(
                params=params)

        # set the range for horizontal rotation
        params['azim_range'] = (
            params['gif_end_azim'] - params['gif_start_azim'])

        # set number of frames for the animated gif as the range of horizontal
        # rotation divided by the number of degrees rotation between frames
        params['steps'] = math.floor(
            params['azim_range']/params['gif_frame_update'])

        # a viewing perspective is composed of an elevation, distance, and
        # azimuth define the range of values we'll cycle through for the
        # distance of the viewing perspective
        params['dist_range'] = np.arange(
            params['gif_min_dist'],
            params['gif_max_dist'],
            (params['gif_max_dist']-params['gif_min_dist'])/params['steps'])

        # define the range of values we'll cycle through for the elevation of
        # the viewing perspective
        params['elev_range'] = np.arange(
            params['gif_max_elev'],
            params['gif_min_elev'],
            (params['gif_min_elev']-params['gif_max_elev'])/params['steps'])

        # now create the individual frames that will be combined later into the
        # animation
        for idx, azimuth in enumerate(
                range(params['gif_start_azim'],
                      params['gif_end_azim'],
                      params['gif_frame_update'])):

            # pan down, rotate around, and zoom out
            ax.azim = float(azimuth)
            ax.elev = params['elev_range'][idx]
            ax.dist = params['dist_range'][idx]

            # set the figure title
            st = fig.suptitle(params['titlename'],
                              fontsize=params['title_font_scale'],
                              fontweight=0,
                              color='black',
                              style='italic',
                              y=1.02)

            st.set_y(0.98)
            fig.subplots_adjust(top=1)

            # save the image as a png file
            plt.savefig('{}/{}/img{:03d}.png'.format(
                params['gif_folder'],
                params['gif_filename'],
                azimuth),
                dpi=params['gif_dpi'])

        # close the image object
        plt.close()

        cls._create_animation(params)


    @staticmethod
    def _gif_defaults_setup(params):

        if params['gif_folder'] is None:
            params['gif_folder'] = params['gif_folder_'+params['graphtype']]
        if params['gif_filename'] is None:
            params['gif_filename'] = params['gif_filename_'+params['graphtype']]

        params['working_folder'] = '{}/{}'.format(
            params['gif_folder'], params['gif_filename'])
        if not os.path.exists(params['working_folder']):
            os.makedirs(params['working_folder'])

        return params


    @staticmethod
    def _create_animation(params):

        # load all the static images into a list then save as an animated gif
        gif_filepath = '{}/{}.gif'.format(
            params['gif_folder'], params['gif_filename'])
        images = ([Image.open(image) for image in sorted(
            glob.glob('{}/*.png'.format(params['working_folder'])))])
        gif = images[0]
        gif.info['duration'] = params['gif_ms'] #milliseconds per frame
        gif.info['loop'] = 0 #how many times to loop (0=infinite)
        gif.save(fp=gif_filepath, format='gif', save_all=True,
                 append_images=images[1:])



"""
Visualize option payoffs and greeks

"""
import copy
from optionvisualizer.visualizer_params import vis_params_dict
from optionvisualizer.option_formulas import Option
from optionvisualizer.simple_payoffs import SimplePayoff
from optionvisualizer.multi_payoffs import MultiPayoff
from optionvisualizer.utilities import Utils
from optionvisualizer.greeks import Greeks
from optionvisualizer.sensitivities import Sens
from optionvisualizer.animated_gifs import Gif
from optionvisualizer.barriers import Barrier

# pylint: disable=invalid-name

class Visualizer():
    """
    Visualize option payoffs and greeks

    """
    def __init__(self, **kwargs):

        # Dictionary of parameter defaults
        self.df_dict = copy.deepcopy(vis_params_dict)

        # Store initial inputs
        inputs = {}
        for key, value in kwargs.items():
            inputs[key] = value

        # Initialise system parameters
        params = Utils._init_params(inputs)

        self.params = params


    def option_data(self, option_value, **kwargs):
        """
        Calculate Option prices or Greeks

        Parameters
        ----------
        option_value : str
            The value to return; price or specified greek
        params : Dict
            S : Float
                Underlying Stock Price. The default is 100.
            K : Float
                Strike Price. The default is 100.
            T : Float
                Time to Maturity. The default is 0.25 (3 months).
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'


        Returns
        -------
        Price / Greek.

        """

        # Update params with the specified parameters
        for key, value in kwargs.items():

            # Replace the default parameter with that provided
            self.params[key] = value

        opt_params = {
            'S':self.params['S'],
            'K':self.params['K'],
            'T':self.params['T'],
            'r':self.params['r'],
            'q':self.params['q'],
            'sigma':self.params['sigma'],
            'option':self.params['option'],
            }

        try:
            # Select the chosen option value from the available functions
            function = self.params['greek_dict'][option_value]
            return getattr(Option, function)(
                opt_params=opt_params, params=self.params)

        except KeyError:
            return print("Please enter a valid function from 'price', "\
                   "'delta', 'gamma', 'vega', 'theta', 'rho', 'vomma', "\
                       "'vanna', 'zomma', 'speed', 'color', 'ultima', "\
                           "'vega bleed', 'charm'")


    def barrier(self, **kwargs):
        """
        Return the Barrier option price

        Parameters
        ----------
        S : Float
            Underlying Stock Price. The default is 100.
        K : Float
            Strike Price. The default is 100.
        H : Float
            Barrier Level. The default is 105.
        R : Float
            Rebate. The default is 0.
        T : Float
            Time to Maturity. The default is 0.25 (3 months).
        r : Float
            Interest Rate. The default is 0.05 (5%).
        q : Float
            Dividend Yield. The default is 0.
        sigma : Float
            Implied Volatility. The default is 0.2 (20%).
        barrier_direction : Str
            Up or Down. The default is 'up'.
        knock : Str
            knock-in or knock-out. The default is 'in'.
        option : Str
            Option type, Put or Call. The default is 'call'


        Returns
        -------
        Float
            Barrier option price.

        """
        # Update params with the specified parameters
        for key, value in kwargs.items():

            # Replace the default parameter with that provided
            self.params[key] = value

        barrier_price, self.params = Barrier.barrier_price(params=self.params)

        return barrier_price


    def sensitivities(self, **kwargs):
        """
        Sensitivities of the option.

        Parameters
        ----------
        S : Float
            Underlying Stock Price. The default is 100.
        K : Float
            Strike Price. The default is 100.
        T : Float
            Time to Maturity. The default is 0.25 (3 months).
        r : Float
            Interest Rate. The default is 0.05 (5%).
        q : Float
            Dividend Yield. The default is 0.
        sigma : Float
            Implied Volatility. The default is 0.2 (20%).
        option : Str
            Option type, Put or Call. The default is 'call'
        greek : Str
            Sensitivity to return. Select from 'delta', 'gamma', 'vega',
            'theta', 'rho', 'vomma', 'vanna', 'zomma', 'speed', 'color',
            'ultima', 'vega bleed', 'charm'. The default is 'delta'
        price_shift : Float
            The size of the price shift in decimal terms. The default
            is 0.25.
        vol_shift : Float
            The size of the volatility shift in decimal terms. The
            default is 0.001.
        ttm_shift : Float
            The size of the time to maturity shift in decimal terms. The
            default is 1/365.
        rate_shift : Float
            The size of the interest rate shift in decimal terms. The
            default is 0.0001.
        num_sens : Bool
            Whether to calculate numerical or analytical sensitivity.
            The default is False.

        Returns
        -------
        Float
            Option Sensitivity.

        """

        # Update params with the specified parameters
        for key, value in kwargs.items():

            # Replace the default parameter with that provided
            self.params[key] = value

        opt_params = {
            'S':self.params['S'],
            'K':self.params['K'],
            'T':self.params['T'],
            'r':self.params['r'],
            'q':self.params['q'],
            'sigma':self.params['sigma'],
            'option':self.params['option'],
            }

        if self.params['num_sens']:
            return Sens.numerical_sensitivities(
                opt_params=opt_params, params=self.params)

        return Sens.analytical_sensitivities(
            opt_params=opt_params, params=self.params)


    def visualize(self, **kwargs):
        """
        Plot the chosen graph of risk or payoff.

        Parameters
        ----------
        risk : Bool
            Whether to display risk graph or payoff graph. The default
            is True.
        S : Float
            Underlying Stock Price. Used in risk & payoff graphs. The
            default is 100.
        T : Float
            Time to Maturity. Used in risk & payoff graphs. The default
            is 0.25 (3 months).
        r : Float
            Interest Rate. Used in risk & payoff graphs. The default
            is 0.05 (5%).
        q : Float
            Dividend Yield. Used in risk & payoff graphs. The default
            is 0.
        sigma : Float
            Implied Volatility. Used in risk & payoff graphs. The default
            is 0.2 (20%).
        option : Str
            Option type, Put or Call. Used in risk & payoff graphs. The
            default is 'call'
        direction : Str
            Whether the payoff is long or short. Used in risk & payoff
            graphs. The default is 'long'.
        greek : Str
            The sensitivity to be charted. Select from 'delta', 'gamma',
            'vega', 'theta', 'rho', 'vomma', 'vanna', 'zomma', 'speed',
            'color', 'ultima', 'vega_bleed', 'charm'. Used in risk
            graphs. The default is 'delta'.
        graphtype : Str
            Whether to plot 2D or 3D graph. Used in risk graphs. The
            default is 2D.
        x_plot : Str
            The x-axis variable - 'price', 'strike', 'vol' or 'time'.
            Used in 2D-risk graphs. The default is 'price'.
        y_plot : Str
            The y-axis variable - 'value', 'delta', 'gamma', 'vega' or
            'theta'. Used in 2D-risk graphs. The default is 'delta.
        G1 : Float
            Strike Price of option 1. Used in 2D-risk graphs. The
            default is 90.
        G2 : Float
            Strike Price of option 2. Used in 2D-risk graphs. The
            default is 100.
        G3 : Float
            Strike Price of option 3. Used in 2D-risk graphs. The
            default is 110.
        T1 : Float
            Time to Maturity of option 1. Used in 2D-risk graphs. The
            default is 0.25 (3 months).
        T2 : Float
            Time to Maturity of option 1. Used in 2D-risk graphs. The
            default is 0.25 (3 months).
        T3 : Float
            Time to Maturity of option 1. Used in 2D-risk graphs. The
            default is 0.25 (3 months).
        time_shift : Float
            Difference between T1 and T2 in rho graphs. Used in 2D-risk
            graphs. The default is 0.25 (3 months).
        interactive : Bool
            Whether to show matplotlib (False) or plotly (True) graph.
            Used in 3D-risk graphs. The default is False.
        notebook : Bool
            Whether the function is being run in an IPython notebook and
            hence whether it should output in line or to an HTML file.
            Used in 3D-risk graphs. The default is False.
        colorscheme : Str
            The matplotlib colormap or plotly colorscale to use. Used
            in 3D-risk graphs. The default is 'jet' (which is a palette
            that works in both plotly and matplotlib).
        colorintensity : Float
            The alpha value indicating level of transparency /
            intensity. The default is 1.
        size2d : Tuple
            Figure size for matplotlib chart. Used in 2D-risk & payoff
            graphs. The default is (6, 4).
        size3d : Tuple
            Figure size for matplotlib chart. Used in 3D-risk graphs.
            The default is (15, 12).
        axis : Str
            Whether the x-axis is 'price' or 'vol'. Used in 3D-risk
            graphs. The default is 'price'.
        spacegrain : Int
            Number of points in each axis linspace argument for 3D
            graphs. Used in 3D-risk graphs. The default is 100.
        azim : Float
            L-R view angle for 3D graphs. The default is -50.
        elev : Float
            Elevation view angle for 3D graphs. The default is 20.
        K : Float
            Strike Price of option 1. Used in payoff graphs. The
            default is 100 (individual payoffs may have own defaults).
        K1 : Float
             Strike Price of option 1. Used in payoff graphs. The
             default is 95 (individual payoffs may have own defaults).
        K2 : Float
             Strike Price of option 2. Used in payoff graphs. The
             default is 105 (individual payoffs may have own defaults).
        K3 : Float
             Strike Price of option 3. Used in payoff graphs. The
             default is 105 (individual payoffs may have own defaults).
        K4 : Float
             Strike Price of option 4. Used in payoff graphs. The
             default is 105 (individual payoffs may have own defaults).
        cash : Bool
            Whether to discount forward to present value. Used in
            forward payoff graph. The default is False.
        ratio : Int
            Multiple of OTM options to be sold for ITM purchased. Used
            in backspread, ratio vertical spread payoff graphs. The
            default is 2.
        value : Bool
            Whether to show the current value as well as the terminal
            payoff. Used in payoff graphs. The default is False.
        combo_payoff : Str
            The payoff to be displayed. Used in payoff graphs. The
            default is 'straddle'.
        mpl_style : Str
            Matplotlib style template for 2D risk charts and payoffs.
            The default is 'seaborn-darkgrid'.
        num_sens : Bool
            Whether to calculate numerical or analytical sensitivity.
            The default is False.
        gif : Bool
            Whether to create an animated gif. The default is False.

        Returns
        -------
        Displays graph of either 2D / 3D greeks or payoff diagram.

        """

        # Update params with the specified parameters
        for key, value in kwargs.items():

            # Replace the default parameter with that provided
            self.params[key] = value

        # If a risk graph is selected
        if self.params['risk']:

            # Run 2D greeks method
            if self.params['graphtype'] == '2D':
                return Greeks.greeks_graphs_2D(params=self.params)

            # Run 3D greeks method
            if self.params['graphtype'] == '3D':
                return Greeks.greeks_graphs_3D(params=self.params)

        else:
            # Specify the combo payoff so that parameter initialisation
            # takes into account specific defaults
            self.params['combo_payoff'] = self.params['payoff_type']

            # Update pricing input parameters to default if not supplied
            self.params = Utils.refresh_combo_params(
                params=self.params, inputs=kwargs)

            return self.payoffs(
                payoff_type=self.params['payoff_type'], params=self.params)


    def greeks(self, **kwargs):
        """
        Plot the chosen 2D or 3D graph

        Parameters
        ----------
        greek : Str
            The sensitivity to be charted. Select from 'delta', 'gamma',
            'vega', 'theta', 'rho', 'vomma', 'vanna', 'zomma', 'speed',
            'color', 'ultima', 'vega_bleed', 'charm'. The default is
            'delta'
        x_plot : Str
                 The x-axis variable ('price', 'strike', 'vol' or
                 'time'). The default is 'price'.
        y_plot : Str
                 The y-axis variable ('value', 'delta', 'gamma', 'vega'
                 or 'theta'). The default is 'delta.
        S : Float
             Underlying Stock Price. The default is 100.
        G1 : Float
             Strike Price of option 1. The default is 90.
        G2 : Float
             Strike Price of option 2. The default is 100.
        G3 : Float
             Strike Price of option 3. The default is 110.
        T : Float
            Time to Maturity. The default is 0.25 (3 months).
        T1 : Float
             Time to Maturity of option 1. The default is 0.25
             (3 months).
        T2 : Float
             Time to Maturity of option 1. The default is 0.25
             (3 months).
        T3 : Float
             Time to Maturity of option 1. The default is 0.25
             (3 months).
        time_shift : Float
             Difference between T1 and T2 in rho graphs. The default
             is 0.25 (3 months).
        r : Float
            Interest Rate. The default is 0.05 (5%).
        q : Float
            Dividend Yield. The default is 0.
        sigma : Float
            Implied Volatility. The default is 0.2 (20%).
        option : Str
            Option type, Put or Call. The default is 'call'
        direction : Str
            Whether the payoff is long or short. The default is 'long'.
        interactive : Bool
            Whether to show matplotlib (False) or plotly (True) graph.
            The default is False.
        notebook : Bool
            Whether the function is being run in an IPython notebook and
            hence whether it should output in line or to an HTML file.
            The default is False.
        colorscheme : Str
            The matplotlib colormap or plotly colorscale to use. The
            default is 'jet' (which works in both).
        colorintensity : Float
            The alpha value indicating level of transparency /
            intensity. The default is 1.
        size2d : Tuple
            Figure size for matplotlib chart. The default is (6, 4).
        size3d : Tuple
            Figure size for matplotlib chart. The default is (15, 12).
        axis : Str
            Whether the x-axis is 'price' or 'vol'. The default is
            'price'.
        spacegrain : Int
            Number of points in each axis linspace argument for 3D
            graphs. The default is 100.
        azim : Float
            L-R view angle for 3D graphs. The default is -50.
        elev : Float
            Elevation view angle for 3D graphs. The default is 20.
        graphtype : Str
            Whether to plot 2D or 3D graph. The default is 2D.
        mpl_style : Str
            Matplotlib style template for 2D risk charts and payoffs.
            The default is 'seaborn-darkgrid'.
        num_sens : Bool
            Whether to calculate numerical or analytical sensitivity.
            The default is False.
        gif : Bool
            Whether to create an animated gif. The default is False.

        Returns
        -------
        Runs method to display either 2D or 3D greeks graph.

        """

        # Update params with the specified parameters
        for key, value in kwargs.items():

            # Replace the default parameter with that provided
            self.params[key] = value

        # Run 2D greeks method
        if self.params['graphtype'] == '2D':
            return Greeks.greeks_graphs_2D(params=self.params)

        # Run 3D greeks method
        elif self.params['graphtype'] == '3D':
            return Greeks.greeks_graphs_3D(params=self.params)

        else:
            print("Please select a '2D' or '3D' graphtype")


    def payoffs(self, payoff_type, **kwargs):
        """
        Displays the graph of the specified combo payoff.

        Parameters
        ----------
        payoff_type : str
            The payoff to be displayed.
        **kwargs : Dict
            S : Float
                 Underlying Stock Price. The default is 100.
            K : Float
                Strike Price of option 1. The default is 100
                (individual payoffs may have own defaults).
            K1 : Float
                 Strike Price of option 1. The default is 95
                 (individual payoffs may have own defaults).
            K2 : Float
                 Strike Price of option 2. The default is 105
                 (individual payoffs may have own defaults).
            K3 : Float
                 Strike Price of option 3. The default is 105
                 (individual payoffs may have own defaults).
            K4 : Float
                 Strike Price of option 4. The default is 105
                 (individual payoffs may have own defaults).
            T : Float
                Time to Maturity. The default is 0.25 (3 months).
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'
            direction : Str
                Whether the payoff is long or short. The default is 'long'.
            cash : Bool
                Whether to discount forward to present value. The default
                is False.
            ratio : Int
                Multiple of OTM options to be sold for ITM purchased. The
                default is 2.
            value : Bool
                Whether to show the current value as well as the terminal
                payoff. The default is False.
            combo_payoff : Str
                The payoff to be displayed.
            mpl_style : Str
                Matplotlib style template for 2D risk charts and payoffs.
                The default is 'seaborn-darkgrid'.

        Returns
        -------
        Runs the specified combo payoff method.

        """
        if 'params' not in kwargs.keys():
            # Update params with the specified parameters
            for key, value in kwargs.items():

                # Replace the default parameter with that provided
                self.params[key] = value

            # Specify the combo payoff so that parameter initialisation
            # takes into account specific defaults
            self.params['combo_payoff'] = payoff_type

            # Update pricing input parameters to default if not supplied
            self.params = Utils.refresh_combo_params(
                params=self.params, inputs=kwargs)

        else:
            self.params = kwargs['params']

        if payoff_type in self.params['combo_simple_dict']:

            # Select the chosen payoff from the available functions
            function = self.params['combo_name_dict'][payoff_type]
            return getattr(SimplePayoff, function)(params=self.params)

        if payoff_type in self.params['combo_multi_dict']:

            # Select the chosen payoff from the available functions
            function = self.params['combo_name_dict'][payoff_type]
            return getattr(MultiPayoff, function)(params=self.params)

        # Otherwise prompt for a valid payoff
        return print("Please enter a valid payoff from 'call', 'put', "\
               "'stock', 'forward', 'collar', 'spread', 'backspread', "\
                   "'ratio vertical spread', 'straddle', 'strangle', "\
                       "'butterfly', 'christmas tree', 'condor', "\
                           "'iron butterfly', 'iron condor'")


    def animated_gif(self, graphtype, **kwargs):
        """
        Create an animated gif of the selected pair of parameters or the
        selected greek 3D graph..

        Parameters
        ----------
        **kwargs : Dict
            gif_folder : Str
                The folder to save the files into. The default is
                'images/greeks'.
            gif_filename : Str
                The filename for the animated gif. The default is 'greek'.
            T : Float
                Time to Maturity. The default is 0.5 (6 months).
            steps : Int
                Number of images to combine. The default is 40.
            x_plot : Str
                     The x-axis variable ('price', 'strike' or 'vol').
                     The default is 'price'.
            y_plot : Str
                     The y-axis variable ('value', 'delta', 'gamma', 'vega'
                     or 'theta'). The default is 'delta.
            S : Float
                 Underlying Stock Price. The default is 100.
            G1 : Float
                 Strike Price of option 1. The default is 90.
            G2 : Float
                 Strike Price of option 2. The default is 100.
            G3 : Float
                 Strike Price of option 3. The default is 110.
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            option : Str
                Option type, Put or Call. The default is 'call'
            direction : Str
                Whether the payoff is long or short. The default is 'long'.
            size2d : Tuple
                Figure size for matplotlib chart. The default is (6, 4).
            mpl_style : Str
                Matplotlib style template for 2D risk charts and payoffs.
                The default is 'seaborn-darkgrid'.
            num_sens : Bool
                Whether to calculate numerical or analytical sensitivity.
                The default is False.
            notebook : Bool
                Whether the function is being run in an IPython notebook and
                hence whether it should output in line or to an HTML file.
                The default is False.
            colorscheme : Str
                The matplotlib colormap or plotly colorscale to use. The
                default is 'jet' (which works in both).
            colorintensity : Float
                The alpha value indicating level of transparency /
                intensity. The default is 1.
            size3d : Tuple
                Figure size for matplotlib chart. The default is (15, 12).
            axis : Str
                Whether the x-axis is 'price' or 'vol'. The default
                is 'price'.
            spacegrain : Int
                Number of points in each axis linspace argument for 3D
                graphs. The default is 100.
            azim : Float
                L-R view angle for 3D graphs. The default is -50.
            elev : Float
                Elevation view angle for 3D graphs. The default is 20.
            greek : Str
                The sensitivity to be charted. Select from 'delta', 'gamma',
                'vega', 'theta', 'rho', 'vomma', 'vanna', 'zomma', 'speed',
                'color', 'ultima', 'vega_bleed', 'charm'. The default is
                'delta'
            gif_folder : Str
                The folder to save the files into. The default is
                'images/greeks'.
            gif_filename : Str
                The filename for the animated gif. The default is 'greek'.
            gif_frame_update : Int
                The number of degrees of rotation between each frame used to
                construct the animated gif. The default is 2.
            gif_min_dist : Float
                The minimum zoom distance. The default is 9.0.
            gif_max_dist : Float
                The maximum zoom distance. The default is 10.0.
            gif_min_elev : Float
                The minimum elevation. The default is 10.0.
            gif_max_elev : Float
                The maximum elevation. The default is 60.0.
            gif_start_azim : Float
                The azimuth to start the gif from. The default is 0.
            gif_end_azim : Float
                The azimuth to end the gif on. The default is 360.
            gif_dpi : Int
                The image resolution to save. The default is 50 dpi.
            gif_ms : Int
                The time to spend on each frame in the gif. The default is
                100ms.

        Returns
        -------
        Saves an animated gif.

        """
        self.params = copy.deepcopy(vis_params_dict)

        self.params['graphtype'] = graphtype
        self.params['gif'] = True

        # Update params with the specified parameters
        for key, value in kwargs.items():

            # Replace the default parameter with that provided
            self.params[key] = value

        # Create a 2D animated gif
        if self.params['graphtype'] == '2D':
            return Gif.animated_2D_gif(params=self.params)

        # Create a 3D animated gif
        if self.params['graphtype'] == '3D':
            return Gif.animated_3D_gif(params=self.params)

        # Otherwise prompt for a valid gif type
        return print("Please select gif_type as '2D' or '3D'")



"""
One and Two Option Payoffs

"""
import numpy as np
from optionvisualizer.multi_payoffs import MultiPayoff
from optionvisualizer.option_formulas import Option
# pylint: disable=invalid-name

class SimplePayoff():
    """
    Calculate One and Two Option payoffs - Call / Put / Spread / Straddle etc.

    """

    @staticmethod
    def call(params):
        """
        Displays the graph of the call.

        Parameters
        ----------
        params : Dict
            S : Float
                 Underlying Stock Price. The default is 100.
            K : Float
                Strike Price of option 1. The default is 100.
            T : Float
                Time to Maturity. The default is 0.25 (3 months).
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            direction : Str
                Whether the payoff is long or short. The default is 'long'.
            value : Bool
                Whether to show the current value as well as the terminal
                payoff. The default is False.
            mpl_style : Str
                Matplotlib style template for 2D risk charts and payoffs.
                The default is 'seaborn-darkgrid'.

        Returns
        -------
        payoff: Terminal payoff value less initial cost; Array
        payoff2: Current payoff value less initial cost (optional); Array
        title: Description of payoff; Str
        Runs method to graph using Matplotlib.

        """

        params['opt_dict'] = {
            'legs':1,
            'S':params['S'],
            'K1':params['K'],
            'T1':params['T'],
            'r':params['r'],
            'q':params['q'],
            'sigma':params['sigma'],
            'option1':'call'
            }

        # Calculate option prices
        option_legs = Option.return_options(
            opt_dict=params['opt_dict'], params=params)

        # Create payoff based on direction
        if params['direction'] == 'long':
            payoff = option_legs['C1'] - option_legs['C1_0']
            title = 'Long Call'
            if params['value']:
                payoff2 = option_legs['C1_G'] - option_legs['C1_0']
            else:
                payoff2 = None

        if params['direction'] == 'short':
            payoff = -option_legs['C1'] + option_legs['C1_0']
            title = 'Short Call'
            if params['value']:
                payoff2 = -option_legs['C1_G'] + option_legs['C1_0']
            else:
                payoff2 = None

        params['payoff_dict'] = {
            'S':params['S'],
            'SA':option_legs['SA'],
            'payoff':payoff,
            'title':title,
            'payoff2':payoff2,
            'size2d':params['size2d'],
            'mpl_style':params['mpl_style']
            }

        # Visualize payoff
        return MultiPayoff.vis_payoff(
            payoff_dict=params['payoff_dict'], params=params)


    @staticmethod
    def put(params):
        """
        Displays the graph of the put.

        Parameters
        ----------
        params : Dict
            S : Float
                 Underlying Stock Price. The default is 100.
            K : Float
                Strike Price of option 1. The default is 100.
            T : Float
                Time to Maturity. The default is 0.25 (3 months).
            r : Float
                Interest Rate. The default is 0.05 (5%).
            q : Float
                Dividend Yield. The default is 0.
            sigma : Float
                Implied Volatility. The default is 0.2 (20%).
            direction : Str
                Whether the payoff is long or short. The default is 'long'.
            value : Bool
                Whether to show the current value as well as the terminal
                payoff. The default is False.
            mpl_style : Str
                Matplotlib style template for 2D risk charts and payoffs.
                The default is 'seaborn-darkgrid'.

        Returns
        -------
        payoff: Terminal payoff value less initial cost; Array
        payoff2: Current payoff value less initial cost (optional); Array
        title: Description of payoff; Str
        Runs method to graph using Matplotlib.

        """

        params['opt_dict'] = {
            'legs':1,
            'S':params['S'],
            'K1':params['K'],
            'T1':params['T'],
            'r':params['r'],
            'q':params['q'],
            'sigma':params['sigma'],
            'option1':'put'
            }

        # Calculate option prices
        option_legs = Option.return_options(
            opt_dict=params['opt_dict'], params=params)

        # Create payoff based on direction
        if params['direction'] == 'long':
            payoff = option_legs['C1'] - option_legs['C1_0']
            title = 'Long Put'
            if params['value']:
                payoff2 = option_legs['C1_G'] - option_legs['C1_0']
            else:
                payoff2 = None

        if params['direction'] == 'short':
            payoff = -option_legs['C1'] + option_legs['C1_0']
            title = 'Short Put'
            if params['value']:
                payoff2 = -option_legs['C1_G'] + option_legs['C1_0']
            else:
                payoff2 = None

        params['payoff_dict'] = {
            'S':params['S'],
            'SA':option_legs['SA'],
            'payoff':payoff,
            'title':title,
            'payoff2':payoff2,
            'size2d':params['size2d'],
            'mpl_style':params['mpl_style']
            }

        # Visualize payoff
        return MultiPayoff.vis_payoff(
            payoff_dict=params['payoff_dict'], params=params)


    @staticmethod
    def stock(params):
        """
        Displays the graph of the underlying.

        Parameters
        ----------
        S : Float
             Underlying Stock Price. The default is 100.
        direction : Str
            Whether the payoff is long or short. The default is 'long'.
        mpl_style : Str
            Matplotlib style template for 2D risk charts and payoffs.
            The default is 'seaborn-darkgrid'.

        Returns
        -------
        payoff: Terminal payoff value less initial cost; Array
        title: Description of payoff; Str
        Runs method to graph using Matplotlib.

        """

        # Define strike range
        SA = np.linspace(0.75 * params['S'], 1.25 * params['S'], 1000)

        # Create payoff based on option type
        if params['direction'] == 'long':
            payoff = SA - params['S']
            title = 'Long Stock'

        if params['direction'] == 'short':
            payoff = params['S'] - SA
            title = 'Short Stock'

        params['payoff_dict'] = {
            'S':params['S'],
            'SA':SA,
            'payoff':payoff,
            'title':title,
            'payoff2':None,
            'size2d':params['size2d'],
            'mpl_style':params['mpl_style']
            }

        # Visualize payoff
        return MultiPayoff.vis_payoff(
            payoff_dict=params['payoff_dict'], params=params)


    @staticmethod
    def forward(params):
        """
        Displays the graph of the synthetic forward strategy:
            Long one ATM call
            Short one ATM put

        Parameters
        ----------
        S : Float
             Underlying Stock Price. The default is 100.
        T : Float
            Time to Maturity. The default is 0.25 (3 months).
        r : Float
            Interest Rate. The default is 0.05 (5%).
        q : Float
            Dividend Yield. The default is 0.
        sigma : Float
            Implied Volatility. The default is 0.2 (20%).
        direction : Str
            Whether the payoff is long or short. The default is 'long'.
        cash : Bool
            Whether to discount to present value. The default is False.
        mpl_style : Str
            Matplotlib style template for 2D risk charts and payoffs.
            The default is 'seaborn-darkgrid'.

        Returns
        -------
        payoff: Terminal payoff value less initial cost; Array
        payoff2: Current payoff value less initial cost (optional); Array
        title: Description of payoff; Str
        Runs method to graph using Matplotlib.

        """

        params['opt_dict'] = {
            'legs':2,
            'S':params['S'],
            'K1':params['S'],
            'T1':params['T'],
            'K2':params['S'],
            'T2':params['T'],
            'r':params['r'],
            'q':params['q'],
            'sigma':params['sigma'],
            'option1':'call',
            'option2':'put'
            }

        # Calculate option prices
        option_legs = Option.return_options(
            opt_dict=params['opt_dict'], params=params)

        # Whether to discount the payoff
        if params['cash']:
            pv = np.exp(-params['r'] * params['T'])
        else:
            pv = 1

        # Create payoff based on option type
        if params['direction'] == 'long':
            payoff = ((option_legs['C1'] - option_legs['C2']
                      - option_legs['C1_0'] + option_legs['C2_0'])
                      * pv)
            title = 'Long Forward'

        if params['direction'] == 'short':
            payoff = ((-option_legs['C1'] + option_legs['C2']
                       + option_legs['C1_0'] - option_legs['C2_0'])
                      * pv)
            title = 'Short Forward'

        params['payoff_dict'] = {
            'S':params['S'],
            'SA':option_legs['SA'],
            'payoff':payoff,
            'title':title,
            'payoff2':None,
            'size2d':params['size2d'],
            'mpl_style':params['mpl_style']
            }

        # Visualize payoff
        return MultiPayoff.vis_payoff(
            payoff_dict=params['payoff_dict'], params=params)


    @staticmethod
    def collar(params):
        """
        Displays the graph of the collar strategy:
            Long one OTM put
            Short one OTM call

        Parameters
        ----------
        S : Float
             Underlying Stock Price. The default is 100.
        K1 : Float
             Strike Price of option 1. The default is 98.
        K2 : Float
             Strike Price of option 2. The default is 102.
        T : Float
            Time to Maturity. The default is 0.25 (3 months).
        r : Float
            Interest Rate. The default is 0.05 (5%).
        q : Float
            Dividend Yield. The default is 0.
        sigma : Float
            Implied Volatility. The default is 0.2 (20%).
        direction : Str
            Whether the payoff is long or short. The default is 'long'.
        value : Bool
            Whether to show the current value as well as the terminal
            payoff. The default is False.
        mpl_style : Str
            Matplotlib style template for 2D risk charts and payoffs.
            The default is 'seaborn-darkgrid'.

        Returns
        -------
        payoff: Terminal payoff value less initial cost; Array
        payoff2: Current payoff value less initial cost (optional); Array
        title: Description of payoff; Str
        Runs method to graph using Matplotlib.

        """

        params['opt_dict'] = {
            'legs':2,
            'S':params['S'],
            'K1':params['K1'],
            'T1':params['T'],
            'K2':params['K2'],
            'T2':params['T'],
            'r':params['r'],
            'q':params['q'],
            'sigma':params['sigma'],
            'option1':'put',
            'option2':'call'
            }

        # Calculate option prices
        option_legs = Option.return_options(
            opt_dict=params['opt_dict'], params=params)

        # Create payoff based on option type
        if params['direction'] == 'long':
            payoff = (option_legs['SA'] - params['S']
                      + option_legs['C1'] - option_legs['C2']
                      - option_legs['C1_0'] + option_legs['C2_0'])
            title = 'Long Collar'

            if params['value']:
                payoff2 = (option_legs['SA'] - params['S']
                           + option_legs['C1_G'] - option_legs['C2_G']
                           - option_legs['C1_0'] + option_legs['C2_0'])
            else:
                payoff2 = None

        if params['direction'] == 'short':
            payoff = (-option_legs['SA'] + params['S']
                      - option_legs['C1'] + option_legs['C2']
                      + option_legs['C1_0'] - option_legs['C2_0'])
            title = 'Short Collar'

            if params['value']:
                payoff2 = (-option_legs['SA'] + params['S']
                           - option_legs['C1_G'] + option_legs['C2_G']
                           + option_legs['C1_0'] - option_legs['C2_0'])
            else:
                payoff2 = None

        params['payoff_dict'] = {
            'S':params['S'],
            'SA':option_legs['SA'],
            'payoff':payoff,
            'title':title,
            'payoff2':payoff2,
            'size2d':params['size2d'],
            'mpl_style':params['mpl_style']
            }

        # Visualize payoff
        return MultiPayoff.vis_payoff(
            payoff_dict=params['payoff_dict'], params=params)


    @staticmethod
    def spread(params):
        """
        Displays the graph of the spread strategy:
            Long one ITM option
            Short one OTM option

        Parameters
        ----------
        S : Float
             Underlying Stock Price. The default is 100.
        K1 : Float
             Strike Price of option 1. The default is 95.
        K2 : Float
             Strike Price of option 2. The default is 105.
        T : Float
            Time to Maturity. The default is 0.25 (3 months).
        r : Float
            Interest Rate. The default is 0.05 (5%).
        q : Float
            Dividend Yield. The default is 0.
        sigma : Float
            Implied Volatility. The default is 0.2 (20%).
        option : Str
            Option type, Put or Call. The default is 'call'
        direction : Str
            Whether the payoff is long or short. The default is 'long'.
        value : Bool
            Whether to show the current value as well as the terminal
            payoff. The default is False.
        mpl_style : Str
            Matplotlib style template for 2D risk charts and payoffs.
            The default is 'seaborn-darkgrid'.

        Returns
        -------
        payoff: Terminal payoff value less initial cost; Array
        payoff2: Current payoff value less initial cost (optional); Array
        title: Description of payoff; Str
        Runs method to graph using Matplotlib.


        """

        params['opt_dict'] = {
            'legs':2,
            'S':params['S'],
            'K1':params['K1'],
            'T1':params['T'],
            'K2':params['K2'],
            'T2':params['T'],
            'r':params['r'],
            'q':params['q'],
            'sigma':params['sigma'],
            'option1':params['option'],
            'option2':params['option']
            }

        # Calculate option prices
        option_legs = Option.return_options(
            opt_dict=params['opt_dict'], params=params)

        # Create payoff based on option type
        if params['direction'] == 'long':
            payoff = (option_legs['C1'] - option_legs['C2']
                      - option_legs['C1_0'] + option_legs['C2_0'])
            if params['value']:
                payoff2 = (option_legs['C1_G'] - option_legs['C2_G']
                           - option_legs['C1_0'] + option_legs['C2_0'])
            else:
                payoff2 = None

        if params['direction'] == 'short':
            payoff = (-option_legs['C1'] + option_legs['C2']
                      + option_legs['C1_0'] - option_legs['C2_0'])
            if params['value']:
                payoff2 = (-option_legs['C1_G'] + option_legs['C2_G']
                           + option_legs['C1_0'] - option_legs['C2_0'])
            else:
                payoff2 = None

        # Create title based on option type and direction
        if params['option'] == 'call' and params['direction'] == 'long':
            title = 'Bull Call Spread'
        if params['option'] == 'put' and params['direction'] == 'long':
            title = 'Bull Put Spread'
        if params['option'] == 'call' and params['direction'] == 'short':
            title = 'Bear Call Spread'
        if params['option'] == 'put' and params['direction'] == 'short':
            title = 'Bear Put Spread'

        params['payoff_dict'] = {
            'S':params['S'],
            'SA':option_legs['SA'],
            'payoff':payoff,
            'title':title,
            'payoff2':payoff2,
            'size2d':params['size2d'],
            'mpl_style':params['mpl_style']
            }

        # Visualize payoff
        return MultiPayoff.vis_payoff(
            payoff_dict=params['payoff_dict'], params=params)


    @staticmethod
    def backspread(params):
        """
        Displays the graph of the backspread strategy:
            Short one ITM option
            Long ratio * OTM options

        Parameters
        ----------
        S : Float
             Underlying Stock Price. The default is 100.
        K1 : Float
             Strike Price of option 1. The default is 95.
        K2 : Float
             Strike Price of option 2. The default is 105.
        T : Float
            Time to Maturity. The default is 0.25 (3 months).
        r : Float
            Interest Rate. The default is 0.05 (5%).
        q : Float
            Dividend Yield. The default is 0.
        sigma : Float
            Implied Volatility. The default is 0.2 (20%).
        option : Str
            Option type, Put or Call. The default is 'call'
        ratio : Int
            Multiple of OTM options to be sold for ITM purchased. The
            default is 2.
        value : Bool
            Whether to show the current value as well as the terminal
            payoff. The default is False.
        mpl_style : Str
            Matplotlib style template for 2D risk charts and payoffs.
            The default is 'seaborn-darkgrid'.

        Returns
        -------
        payoff: Terminal payoff value less initial cost; Array
        payoff2: Current payoff value less initial cost (optional); Array
        title: Description of payoff; Str
        Runs method to graph using Matplotlib.

        """

        params['opt_dict'] = {
            'legs':2,
            'S':params['S'],
            'K1':params['K1'],
            'T1':params['T'],
            'K2':params['K2'],
            'T2':params['T'],
            'r':params['r'],
            'q':params['q'],
            'sigma':params['sigma'],
            'option1':params['option'],
            'option2':params['option']
            }

        # Calculate option prices
        option_legs = Option.return_options(
            opt_dict=params['opt_dict'], params=params)

        # Create payoff based on option type
        if params['option'] == 'call':
            title = 'Call Backspread'
            payoff = (-option_legs['C1'] + (params['ratio']
                                            * option_legs['C2'])
                      + option_legs['C1_0'] - (params['ratio']
                                               * option_legs['C2_0']))
            if params['value']:
                payoff2 = (-option_legs['C1_G'] + (params['ratio']
                                                   * option_legs['C2_G'])
                           + option_legs['C1_0'] - (params['ratio']
                                                    * option_legs['C2_0']))
            else:
                payoff2 = None

        if params['option'] == 'put':
            payoff = (params['ratio'] * option_legs['C1']
                      - option_legs['C2']
                      - params['ratio'] * option_legs['C1_0']
                      + option_legs['C2_0'])
            title = 'Put Backspread'
            if params['value']:
                payoff2 = (params['ratio'] * option_legs['C1_G']
                           - option_legs['C2_G']
                           - params['ratio'] * option_legs['C1_0']
                           + option_legs['C2_0'])
            else:
                payoff2 = None

        params['payoff_dict'] = {
            'S':params['S'],
            'SA':option_legs['SA'],
            'payoff':payoff,
            'title':title,
            'payoff2':payoff2,
            'size2d':params['size2d'],
            'mpl_style':params['mpl_style']
            }

        # Visualize payoff
        return MultiPayoff.vis_payoff(
            payoff_dict=params['payoff_dict'], params=params)


    @staticmethod
    def ratio_vertical_spread(params):
        """
        Displays the graph of the ratio vertical spread strategy:
            Long one ITM option
            Short ratio * OTM options

        Parameters
        ----------
        S : Float
             Underlying Stock Price. The default is 100.
        K1 : Float
             Strike Price of option 1. The default is 95.
        K2 : Float
             Strike Price of option 2. The default is 105.
        T : Float
            Time to Maturity. The default is 0.25 (3 months).
        r : Float
            Interest Rate. The default is 0.05 (5%).
        q : Float
            Dividend Yield. The default is 0.
        sigma : Float
            Implied Volatility. The default is 0.2 (20%).
        option : Str
            Option type, Put or Call. The default is 'call'
        ratio : Int
            Multiple of OTM options to be sold for ITM purchased. The
            default is 2.
        direction : Str
            Whether the payoff is long or short. The default is 'long'.
        value : Bool
            Whether to show the current value as well as the terminal
            payoff. The default is False.
        mpl_style : Str
            Matplotlib style template for 2D risk charts and payoffs.
            The default is 'seaborn-darkgrid'.

        Returns
        -------
        payoff: Terminal payoff value less initial cost; Array
        payoff2: Current payoff value less initial cost (optional); Array
        title: Description of payoff; Str
        Runs method to graph using Matplotlib.

        """

        params['opt_dict'] = {
            'legs':2,
            'S':params['S'],
            'K1':params['K1'],
            'T1':params['T'],
            'K2':params['K2'],
            'T2':params['T'],
            'r':params['r'],
            'q':params['q'],
            'sigma':params['sigma'],
            'option1':params['option'],
            'option2':params['option']
            }

        # Calculate option prices
        option_legs = Option.return_options(
            opt_dict=params['opt_dict'], params=params)

        # Create payoff based on option type
        if params['option'] == 'call':
            title = 'Call Ratio Vertical Spread'
            payoff = (option_legs['C1']
                      - params['ratio'] * option_legs['C2']
                      - option_legs['C1_0']
                      + params['ratio'] * option_legs['C2_0'])
            if params['value']:
                payoff2 = (option_legs['C1_G']
                           - params['ratio'] * option_legs['C2_G']
                           - option_legs['C1_0']
                           + params['ratio'] * option_legs['C2_0'])
            else:
                payoff2 = None

        if params['option'] == 'put':
            title = 'Put Ratio Vertical Spread'
            payoff = (-params['ratio'] * option_legs['C1']
                      + option_legs['C2']
                      + params['ratio'] * option_legs['C1_0']
                      - option_legs['C2_0'])
            if params['value']:
                payoff2 = (-params['ratio'] * option_legs['C1_G']
                           + option_legs['C2_G']
                           + params['ratio'] * option_legs['C1_0']
                           - option_legs['C2_0'])
            else:
                payoff2 = None

        params['payoff_dict'] = {
            'S':params['S'],
            'SA':option_legs['SA'],
            'payoff':payoff,
            'title':title,
            'payoff2':payoff2,
            'size2d':params['size2d'],
            'mpl_style':params['mpl_style']
            }

        # Visualize payoff
        return MultiPayoff.vis_payoff(
            payoff_dict=params['payoff_dict'], params=params)


    @staticmethod
    def straddle(params):
        """
        Displays the graph of the straddle strategy:
            Long one ATM put
            Long one ATM call

        Parameters
        ----------
        S : Float
             Underlying Stock Price. The default is 100.
        K : Float
            Strike Price of options 1 and 2. The default is 100.
        T : Float
            Time to Maturity. The default is 0.25 (3 months).
        r : Float
            Interest Rate. The default is 0.05 (5%).
        q : Float
            Dividend Yield. The default is 0.
        sigma : Float
            Implied Volatility. The default is 0.2 (20%).
        direction : Str
            Whether the payoff is long or short. The default is 'long'.
        value : Bool
            Whether to show the current value as well as the terminal
            payoff. The default is False.
        mpl_style : Str
            Matplotlib style template for 2D risk charts and payoffs.
            The default is 'seaborn-darkgrid'.

        Returns
        -------
        payoff: Terminal payoff value less initial cost; Array
        payoff2: Current payoff value less initial cost (optional); Array
        title: Description of payoff; Str
        Runs method to graph using Matplotlib.

        """

        params['opt_dict'] = {
            'legs':2,
            'S':params['S'],
            'K1':params['K'],
            'T1':params['T'],
            'K2':params['K'],
            'T2':params['T'],
            'r':params['r'],
            'q':params['q'],
            'sigma':params['sigma'],
            'option1':'put',
            'option2':'call'
            }

        # Calculate option prices
        option_legs = Option.return_options(
            opt_dict=params['opt_dict'], params=params)

        # Create payoff based on direction
        if params['direction'] == 'long':
            payoff = (option_legs['C1'] + option_legs['C2']
                      - option_legs['C1_0'] - option_legs['C2_0'])
            title = 'Long Straddle'
            if params['value']:
                payoff2 = (option_legs['C1_G'] + option_legs['C2_G']
                           - option_legs['C1_0'] - option_legs['C2_0'])
            else:
                payoff2 = None

        if params['direction'] == 'short':
            payoff = (-option_legs['C1'] - option_legs['C2']
                      + option_legs['C1_0'] + option_legs['C2_0'])
            title = 'Short Straddle'
            if params['value']:
                payoff2 = (-option_legs['C1_G'] - option_legs['C2_G']
                           + option_legs['C1_0'] + option_legs['C2_0'])
            else:
                payoff2 = None

        params['payoff_dict'] = {
            'S':params['S'],
            'SA':option_legs['SA'],
            'payoff':payoff,
            'title':title,
            'payoff2':payoff2,
            'size2d':params['size2d'],
            'mpl_style':params['mpl_style']
            }

        # Visualize payoff
        return MultiPayoff.vis_payoff(
            payoff_dict=params['payoff_dict'], params=params)


    @staticmethod
    def strangle(params):
        """
        Displays the graph of the strangle strategy:
            Long one OTM put
            Long one OTM call

        Parameters
        ----------
        S : Float
             Underlying Stock Price. The default is 100.
        K1 : Float
             Strike Price of option 1. The default is 95.
        K2 : Float
             Strike Price of option 2. The default is 105.
        T : Float
            Time to Maturity. The default is 0.25 (3 months).
        r : Float
            Interest Rate. The default is 0.05 (5%).
        q : Float
            Dividend Yield. The default is 0.
        sigma : Float
            Implied Volatility. The default is 0.2 (20%).
        direction : Str
            Whether the payoff is long or short. The default is 'long'.
        value : Bool
            Whether to show the current value as well as the terminal
            payoff. The default is False.
        mpl_style : Str
            Matplotlib style template for 2D risk charts and payoffs.
            The default is 'seaborn-darkgrid'.

        Returns
        -------
        payoff: Terminal payoff value less initial cost; Array
        payoff2: Current payoff value less initial cost (optional); Array
        title: Description of payoff; Str
        Runs method to graph using Matplotlib.

        """

        params['opt_dict'] = {
            'legs':2,
            'S':params['S'],
            'K1':params['K1'],
            'T1':params['T'],
            'K2':params['K2'],
            'T2':params['T'],
            'r':params['r'],
            'q':params['q'],
            'sigma':params['sigma'],
            'option1':'put',
            'option2':'call'
            }

        # Calculate option prices
        option_legs = Option.return_options(
            opt_dict=params['opt_dict'], params=params)

        # Create payoff based on direction
        if params['direction'] == 'long':
            payoff = (option_legs['C1'] + option_legs['C2']
                      - option_legs['C1_0'] - option_legs['C2_0'])
            title = 'Long Strangle'
            if params['value']:
                payoff2 = (option_legs['C1_G'] + option_legs['C2_G']
                           - option_legs['C1_0'] - option_legs['C2_0'])
            else:
                payoff2 = None

        if params['direction'] == 'short':
            payoff = (-option_legs['C1'] - option_legs['C2']
                      + option_legs['C1_0'] + option_legs['C2_0'])
            title = 'Short Strangle'
            if params['value']:
                payoff2 = (-option_legs['C1_G'] - option_legs['C2_G']
                           + option_legs['C1_0'] + option_legs['C2_0'])
            else:
                payoff2 = None

        params['payoff_dict'] = {
            'S':params['S'],
            'SA':option_legs['SA'],
            'payoff':payoff,
            'title':title,
            'payoff2':payoff2,
            'size2d':params['size2d'],
            'mpl_style':params['mpl_style']
            }

        # Visualize payoff
        return MultiPayoff.vis_payoff(
            payoff_dict=params['payoff_dict'], params=params)
